from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, overload

import numpy as np
from cmap import Color, Colormap
from numpy.typing import ArrayLike, NDArray
from psygnal import Signal

from whitecanvas import theme
from whitecanvas.backend import Backend
from whitecanvas.layers._base import DataBoundLayer, LayerEvents
from whitecanvas.protocols import ImageProtocol
from whitecanvas.types import (
    ArrayLike1D,
    ColormapType,
    ColorType,
    HistBinType,
    KdeBandWidthType,
    Orientation,
    OrientationLike,
    Origin,
    Rect,
    _Void,
)
from whitecanvas.utils.normalize import as_array_1d
from whitecanvas.utils.type_check import is_real_number

if TYPE_CHECKING:
    from whitecanvas.layers import Texts, _mixin
    from whitecanvas.layers.group import Colorbar, LabeledImage

    MultiFontTexts = Texts[_mixin.MonoFace, _mixin.MonoEdge, _mixin.MultiFont]

_void = _Void()


class ImageEvents(LayerEvents):
    cmap = Signal(Colormap)
    clim = Signal(tuple)
    shift = Signal(tuple)
    scale = Signal(tuple)


class Image(DataBoundLayer[ImageProtocol, NDArray[np.number]]):
    """
    Grayscale or RGBA image layer.

    Parameters
    ----------
    image : array_like
        2D or 3D array of image data.
    cmap : colormap type, default "gray"
        Colormap to use.
    clim : tuple of float or None, optional
        Contrast limits. If `None`, the limits are set to the min and max of the data.
        You can also pass `None` separately to either limit to only autoscale one of
        them.
    origin : str or Origin, default "corner"
        Origin of the image. This is a redundant parameter which overlaps with `shift`,
        but it makes it easier to operate on the image.
    """

    events: ImageEvents
    _events_class = ImageEvents
    _NO_PADDING_NEEDED = True

    def __init__(
        self,
        image: ArrayLike,
        *,
        name: str | None = None,
        cmap: ColormapType = "gray",
        clim: tuple[float | None, float | None] | None = None,
        shift: tuple[float, float] = (0, 0),
        scale: tuple[float, float] = (1.0, 1.0),
        backend: Backend | str | None = None,
    ):
        img = _normalize_image(image)
        self._origin = Origin.CORNER
        super().__init__(name=name)
        self._backend = self._create_backend(Backend(backend), img)
        if img.ndim == 3:
            cmap = clim = _void
        self._x_hint = self._y_hint = None
        self.update(cmap=cmap, clim=clim, shift=shift, scale=scale)

    def _get_layer_data(self) -> NDArray[np.number]:
        """Current image data of the layer."""
        return self._backend._plt_get_data()

    def _norm_layer_data(self, data: Any) -> NDArray[np.number]:
        return _normalize_image(data)

    def _set_layer_data(self, data: NDArray[np.number]):
        """Set the data of the layer."""
        self._backend._plt_set_data(data)

    @property
    def cmap(self) -> Colormap:
        """Current colormap."""
        return self._backend._plt_get_colormap()

    @cmap.setter
    def cmap(self, cmap: ColormapType):
        _cmap = Colormap(cmap)
        self._backend._plt_set_colormap(_cmap)
        self.events.cmap.emit(_cmap)

    @property
    def clim(self) -> tuple[float, float]:
        """Current contrast limits."""
        return self._backend._plt_get_clim()

    @clim.setter
    def clim(self, clim: tuple[float | None, float | None] | None):
        if clim is None:
            low, high = None, None
        else:
            low, high = clim
        if low is None:
            low = self.data.min()
        if high is None:
            high = self.data.max()
        self._backend._plt_set_clim((low, high))
        self.events.clim.emit((low, high))

    @property
    def shape(self) -> tuple[int, int]:
        """The visual shape of the image (shape without the color axis)."""
        if self.is_rgba:
            return self.data.shape[:2]
        else:
            return self.data.shape

    @property
    def data_mapped(self) -> NDArray[np.number]:
        """The colored image data (N, M, 4) mapped by the colormap."""
        if self.is_rgba:
            return self.data
        # normalize data to [0, 1]
        cmin, cmax = self.clim
        data_norm = (self.data - cmin) / (cmax - cmin)
        return self.cmap(data_norm)

    @property
    def shift(self) -> tuple[float, float]:
        """Current shift from the origin."""
        shift = self._backend._plt_get_translation()
        sx, sy = self.scale
        if self.origin is Origin.EDGE:
            shift = shift[0] - 0.5 * sx, shift[1] - 0.5 * sy
        elif self.origin is Origin.CORNER:
            pass
        elif self.origin is Origin.CENTER:
            sizex, sizey = self.data.shape[:2]
            shift = shift[0] + (sizex - 1) / 2 * sx, shift[1] + (sizey - 1) / 2 * sy
        else:
            raise RuntimeError("Unreachable")
        return shift

    @shift.setter
    def shift(self, shift: tuple[float, float]):
        img = self.data
        sx, sy = self.scale
        if self.origin is Origin.EDGE:
            shift = shift[0] + 0.5 * sx, shift[1] + 0.5 * sy
        elif self.origin is Origin.CORNER:
            pass
        elif self.origin is Origin.CENTER:
            sizex, sizey = img.shape[:2]
            shift = shift[0] - (sizex - 1) / 2 * sx, shift[1] - (sizey - 1) / 2 * sy
        else:
            raise RuntimeError("Unreachable")
        self._backend._plt_set_translation(shift)
        self._x_hint, self._y_hint = _hint_for(
            (img.shape[1], img.shape[0]), shift=shift, scale=self.scale
        )
        self.events.shift.emit(shift)

    @property
    def shift_raw(self) -> tuple[float, float]:
        """Current shift from the origin as a raw data."""
        return self._backend._plt_get_translation()

    @property
    def scale(self) -> tuple[float, float]:
        """Current scale."""
        return self._backend._plt_get_scale()

    @scale.setter
    def scale(self, scale: float | tuple[float, float]):
        if is_real_number(scale):
            scale = float(scale), float(scale)
        dx, dy = scale
        if dx <= 0 or dy <= 0:
            raise ValueError("Scale must be positive.")
        self._backend._plt_set_scale(scale)
        shape = self.data.shape[:2]
        self._x_hint, self._y_hint = _hint_for(
            (shape[1], shape[0]), shift=self.shift, scale=scale
        )
        self.events.scale.emit(scale)

    @property
    def origin(self) -> Origin:
        """Current origin of the image."""
        return self._origin

    @origin.setter
    def origin(self, origin: Origin | str):
        self._origin = Origin(origin)
        self.shift = self.shift  # recalculate

    def update(
        self,
        *,
        cmap: ColormapType | _Void = _void,
        clim: tuple[float | None, float | None] | None | _Void = _void,
        shift: tuple[float, float] | _Void = _void,
        scale: tuple[float, float] | _Void = _void,
        origin: str | Origin | _Void = _void,
    ) -> Image:
        if cmap is not _void:
            if self.is_rgba:
                raise ValueError("Cannot set colormap for an RGBA image.")
            self.cmap = cmap
        if clim is not _void:
            if self.is_rgba:
                raise ValueError("Cannot set contrast limits for an RGBA image.")
            self.clim = clim
        if origin is not _void:
            self.origin = origin
        if shift is not _void:
            self.shift = shift
        if scale is not _void:
            self.scale = scale
        return self

    @overload
    def fit_to(self, bbox: Rect | tuple[float, float, float, float], /) -> Image:
        ...

    @overload
    def fit_to(self, left: float, right: float, bottom: float, top: float, /) -> Image:
        ...

    def fit_to(self, *args) -> Image:
        """Fit the image to the given bounding box."""
        if len(args) == 1:
            rect = Rect(*args[0])
        elif len(args) == 4:
            rect = Rect(*args)
        else:
            raise TypeError("fit_to() takes 1 or 4 positional arguments.")
        shape = self.data.shape[:2]
        x0, y0 = rect.left, rect.bottom
        sx, sy = rect.width / shape[1], rect.height / shape[0]
        if self.origin is Origin.CORNER:
            self.shift = (x0 + 0.5 * sx, y0 + 0.5 * sy)
        elif self.origin is Origin.EDGE:
            self.shift = x0, y0
        else:
            ny, nx = shape
            self.shift = x0 + (nx + 1) / 2 * sx, y0 + (ny + 1) / 2 * sy
        self.scale = sx, sy
        return self

    def with_text(
        self,
        *,
        size: int = 8,
        color_rule: ColorType | Callable[[np.ndarray], ColorType] | None = None,
        fmt: str = "",
        text_invalid: str | None = None,
        mask: NDArray[np.bool_] | None = None,
    ) -> LabeledImage:
        """
        Add text layer that displays the pixel values of the image.

        Parameters
        ----------
        size : int, default 8
            Font size of the text.
        color_rule : color-like, callable, optional
            Rule to define the color for each text based on the color-mapped image
            intensity.
        fmt : str, optional
            Format string for the text.
        mask : array_like, optional
            Mask to specify which pixel to add text if specified.
        """
        from whitecanvas.layers.group import LabeledImage

        text_layer = self._make_text_layer(
            size=size, color_rule=color_rule, fmt=fmt, text_invalid=text_invalid,
            mask=mask,
        )  # fmt: skip
        return LabeledImage(self, texts=text_layer, name=self.name)

    def with_colorbar(
        self,
        bbox: Rect | None = None,
        *,
        orient: OrientationLike = "vertical",
    ) -> LabeledImage:
        """
        Add a colorbar to the image.

        Parameters
        ----------
        bbox : four float, optional
            Bounding box of the colorbar. If `None`, the colorbar will be placed at the
            bottom-right corner of the image.
        orient : str or Orientation, default "vertical"
            Orientation of the colorbar.
        """
        from whitecanvas.layers.group import LabeledImage

        ori = Orientation.parse(orient)
        cbar = self._make_colorbar(bbox=bbox, orient=ori)
        if canvas := self._canvas_ref():
            if ori.is_vertical and canvas.y.flipped:
                cbar.lut.data = cbar.lut.data[::-1]
            elif ori.is_horizontal and canvas.x.flipped:
                cbar.lut.data = cbar.lut.data[:, ::-1]
        return LabeledImage(self, colorbar=cbar, name=self.name)

    def _make_text_layer(
        self,
        *,
        size: int = 8,
        color_rule: ColorType | Callable[[np.ndarray], ColorType] | None = None,
        fmt: str = "",
        text_invalid: str | None = None,
        mask: NDArray[np.bool_] | None = None,
    ) -> MultiFontTexts:
        from whitecanvas.layers._primitive import Texts

        # normalize color_rule
        _color_rule: Callable[[np.ndarray], np.ndarray]

        def _norm_color(x) -> NDArray[np.float32]:
            if isinstance(x, np.ndarray):
                return x.astype(np.float32, copy=False)
            return np.array(Color(x).rgba, dtype=np.float32)

        if color_rule is None:
            _b = _norm_color("black")
            _w = _norm_color("white")
            _bg = _norm_color(theme.get_theme().background_color)[:3]

            def _color_rule(x: NDArray[np.number]) -> NDArray[np.float32]:
                alpha = x[3]
                _col = x[:3] * alpha + _bg * (1 - alpha)
                return _b if _col.sum() > 1.5 else _w

        elif callable(color_rule):
            _color_rule = lambda x: _norm_color(color_rule(x))  # noqa: E731
        else:
            _col = _norm_color(color_rule)
            _color_rule = lambda _: _col  # noqa: E731

        # normalize mask
        if mask is None:
            ij_iter = np.ndindex(self.data.shape[:2])
        else:
            mask = np.asarray(mask, dtype=bool)
            if mask.shape != self.data.shape[:2]:
                raise ValueError(
                    f"Mask shape {mask.shape} must be the same as the image "
                    f"shape {self.data.shape[:2]}."
                )
            ij_iter = np.argwhere(mask)

        img_data = self.data
        img_color = self.data_mapped
        ny, nx = self.shape
        dx, dy = self.shift_raw
        sx, sy = self.scale
        ys = np.arange(ny) * sy + dy
        xs = np.arange(nx) * sx + dx
        texts: list[str] = []
        xdata: list[float] = []
        ydata: list[float] = []
        colors: list[np.ndarray] = []
        # normalize fmt
        if fmt:
            if fmt.startswith(":"):
                fmt_style = "{" + fmt + "}"
            elif fmt.startswith("{") and fmt.endswith("}"):
                fmt_style = fmt
            else:
                fmt_style = "{:" + fmt + "}"
        else:
            fmt_style = "{}"

        for iy, ix in ij_iter:
            x = xs[ix]
            y = ys[iy]
            if np.isfinite(img_data[iy, ix]):
                text = fmt_style.format(img_data[iy, ix])
            else:
                if text_invalid is None:
                    text = repr(img_data[iy, ix])
                else:
                    text = text_invalid
            texts.append(text)
            xdata.append(x)
            ydata.append(y)
            colors.append(_color_rule(img_color[iy, ix]))
        return Texts(
            np.array(xdata), np.array(ydata), texts, size=size, anchor="center"
        ).with_font_multi(color=np.stack(colors, axis=0))

    def _make_colorbar(
        self,
        bbox: Rect | None = None,
        orient: OrientationLike = "vertical",
    ) -> Colorbar:
        from whitecanvas.layers.group import Colorbar

        orient = Orientation.parse(orient)
        cbar = Colorbar(self.cmap, name=f"colorbar<{self.name}>", orient=orient)
        if bbox is None:
            img_bbox = self.bbox
            if orient.is_vertical:
                bbox = Rect(
                    img_bbox.left + img_bbox.width * 0.92,
                    img_bbox.left + img_bbox.width * 0.97,
                    img_bbox.bottom + img_bbox.height * 0.03,
                    img_bbox.bottom + img_bbox.height * 0.23,
                )
            else:
                bbox = Rect(
                    img_bbox.left + img_bbox.width * 0.77,
                    img_bbox.left + img_bbox.width * 0.97,
                    img_bbox.bottom + img_bbox.height * 0.03,
                    img_bbox.bottom + img_bbox.height * 0.08,
                )
        cbar.fit_to(bbox)
        return cbar

    @property
    def is_rgba(self) -> bool:
        """Whether the image is RGBA."""
        return self.data.ndim == 3

    @property
    def bbox(self) -> Rect:
        """Bounding box of the image."""
        ox, oy = self.shift_raw
        sizey, sizex = self.data.shape[:2]
        sx, sy = self.scale
        return Rect(
            left=ox - 0.5 * sx,
            right=ox + (sizex - 0.5) * sx,
            bottom=oy - 0.5 * sy,
            top=oy + (sizey - 0.5) * sy,
        )

    @classmethod
    def build_hist(
        cls,
        x: ArrayLike1D,
        y: ArrayLike1D,
        bins: HistBinType | tuple[HistBinType, HistBinType] = "auto",
        range=None,
        name: str | None = None,
        density: bool = False,
        cmap: ColormapType = "inferno",
        backend: Backend | str | None = None,
    ):
        _x = as_array_1d(x)
        _y = as_array_1d(y)
        if _x.size != _y.size:
            raise ValueError("x and y must have the same size.")
        if isinstance(bins, tuple):
            xbins, ybins = bins
        else:
            xbins = ybins = bins
        if range is None:
            xrange = yrange = None
        else:
            xrange, yrange = range
        _bins = (
            np.histogram_bin_edges(_x, xbins, xrange),
            np.histogram_bin_edges(_y, ybins, yrange),
        )
        hist, xedges, yedges = np.histogram2d(
            _x, _y, bins=_bins, range=(xrange, yrange), density=density
        )
        hist_t = hist.T
        shift = (xedges[0], yedges[0])
        scale = (xedges[1] - xedges[0], yedges[1] - yedges[0])
        self = cls(
            hist_t, name=name, cmap=cmap, shift=shift, scale=scale, backend=backend
        )
        self.origin = Origin.EDGE
        return self

    @classmethod
    def build_kde(
        cls,
        x: ArrayLike1D,
        y: ArrayLike1D,
        shape: tuple[int, int] = (256, 256),
        range=None,
        band_width: KdeBandWidthType = "scott",
        name: str | None = None,
        cmap: ColormapType = "inferno",
        backend: Backend | str | None = None,
    ) -> Image:
        from whitecanvas.utils.kde import gaussian_kde

        _x = as_array_1d(x)
        _y = as_array_1d(y)
        kde = gaussian_kde([_x, _y], bw_method=band_width)
        if range is None:
            xrange = yrange = None
        else:
            xrange, yrange = range
        if xrange is None:
            xrange = _x.min(), _x.max()
        if yrange is None:
            yrange = _y.min(), _y.max()
        xedges = np.linspace(*xrange, shape[0])
        yedges = np.linspace(*yrange, shape[1])
        xx, yy = np.meshgrid(xedges, yedges)
        positions = np.vstack([xx.ravel(), yy.ravel()])
        val = np.reshape(kde(positions).T, xx.shape)
        shift = (xedges[0], yedges[0])
        scale = (xedges[1] - xedges[0], yedges[1] - yedges[0])
        self = cls(val, name=name, cmap=cmap, shift=shift, scale=scale, backend=backend)
        self.origin = Origin.EDGE
        return self


def _normalize_image(image):
    img = np.asarray(image)
    if img.dtype.kind not in "uif":
        raise TypeError(f"Only numerical arrays are allowed, got {img.dtype}")
    # check shape
    if img.ndim == 2:
        pass
    elif img.ndim == 3:
        nchannels = img.shape[2]
        if nchannels not in (3, 4):
            raise ValueError(
                "If 3D array is given, the last dimension must be 3 or 4, "
                f"got shape {img.shape}."
            )
    else:
        raise ValueError(f"Only 2D or 3D arrays are allowed, got {img.ndim}")
    return img


def _hint_for(
    shape: tuple[int, int],
    shift: tuple[float, float] = (0, 0),
    scale: tuple[float, float] = (1, 1),
) -> tuple[float, float]:
    xhint = np.array([-0.5, shape[0] - 0.5]) * scale[0] + shift[0]
    yhint = np.array([-0.5, shape[1] - 0.5]) * scale[1] + shift[1]
    return tuple(xhint), tuple(yhint)



from __future__ import annotations

import bokeh.models as bk_models
import numpy as np

from whitecanvas.backend.bokeh._base import (
    BokehLayer,
    SupportsMouseEvents,
    from_bokeh_hatch,
    to_bokeh_hatch,
)
from whitecanvas.protocols import BandProtocol, check_protocol
from whitecanvas.types import Hatch, LineStyle, Orientation
from whitecanvas.utils.normalize import arr_color, hex_color


@check_protocol(BandProtocol)
class Band(BokehLayer["bk_models.VArea | bk_models.HArea"], SupportsMouseEvents):
    def __init__(
        self,
        t: np.ndarray,
        ydata0: np.ndarray,
        ydata1: np.ndarray,
        orient: Orientation,
    ):
        self._data = bk_models.ColumnDataSource(
            data={
                "t": t,
                "y0": ydata0,
                "y1": ydata1,
                "hovertexts": np.array([""] * len(t)),
            }
        )
        if orient.is_vertical:
            self._model = bk_models.VArea(x="t", y1="y0", y2="y1")
        else:
            self._model = bk_models.HArea(y="t", x1="y0", x2="y1")
        self._edge_color = np.zeros(4)
        self._edge_width = 0
        self._edge_style = LineStyle.SOLID
        self._visible = True
        self._face_color = self._model.fill_color

    def _plt_get_visible(self):
        return self._visible

    def _plt_set_visible(self, visible):
        if visible:
            self._model.fill_color = "#00000000"
        else:
            self._model.fill_color = self._face_color
        self._visible = visible

    ##### XYYDataProtocol #####
    def _plt_get_vertical_data(self):
        data = self._data.data
        return data["t"], data["y0"], data["y1"]

    _plt_get_horizontal_data = _plt_get_vertical_data

    def _plt_set_vertical_data(self, t, ydata0, ydata1):
        cur_data = self._data.data.copy()
        cur_data.update({"t": t, "y0": ydata0, "y1": ydata1})
        self._data.data = cur_data

    _plt_set_horizontal_data = _plt_set_vertical_data

    def _plt_get_face_color(self):
        return arr_color(self._face_color)

    def _plt_set_face_color(self, color):
        self._face_color = hex_color(color)
        if self._visible:
            self._model.fill_color = self._face_color

    def _plt_get_face_hatch(self) -> Hatch:
        return from_bokeh_hatch(self._model.hatch_pattern)

    def _plt_set_face_hatch(self, pattern: Hatch):
        self._model.hatch_pattern = to_bokeh_hatch(pattern)

    def _plt_get_edge_color(self):
        return self._edge_color

    def _plt_set_edge_color(self, color):
        self._edge_color = arr_color(color)

    def _plt_get_edge_width(self):
        return self._edge_width

    def _plt_set_edge_width(self, width: float):
        self._edge_width = width

    def _plt_get_edge_style(self):
        return self._edge_style

    def _plt_set_edge_style(self, style: LineStyle):
        self._edge_style = style

    def _plt_set_hover_text(self, text: str):
        self._data.data["hovertexts"] = [text] * len(self._data.data["t"])



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Iterable, Sequence

import numpy as np
from numpy.typing import NDArray

from whitecanvas import theme
from whitecanvas.backend import Backend
from whitecanvas.layers._primitive import Line, Markers
from whitecanvas.layers.group._collections import LayerCollectionBase
from whitecanvas.types import Hatch, LineStyle, Symbol, XYData
from whitecanvas.utils.normalize import as_any_1d_array, as_color_array, parse_texts
from whitecanvas.utils.type_check import is_real_number

if TYPE_CHECKING:
    from typing_extensions import Self


class LineCollection(LayerCollectionBase[Line]):
    """Collection of lines."""

    def __init__(
        self,
        segments: list[Any],
        name: str | None = None,
        backend: str | Backend | None = None,
    ):
        lines = [Line([], [], backend=backend) for _ in segments]
        for line, seg in zip(lines, segments):
            line.data = seg
        super().__init__(lines, name=name)

    @property
    def data(self) -> list[XYData]:
        return [line.data for line in self]

    @data.setter
    def data(self, data: list[XYData]):
        ndata_in = len(data)
        ndata_now = len(self)
        if ndata_in > ndata_now:
            for _ in range(ndata_now, ndata_in):
                self.append(Line([], []))
        elif ndata_in < ndata_now:
            for _ in range(ndata_in, ndata_now):
                del self[-1]
        for line, d in zip(self, data):
            line.data = d

    @property
    def width(self) -> NDArray[np.float32]:
        """Array of line widths."""
        return np.array([line.width for line in self], dtype=np.float32)

    @width.setter
    def width(self, width: float | Sequence[float]):
        if is_real_number(width):
            _width = [width] * len(self)
        else:
            _width = np.asarray(width, dtype=np.float32)
        if len(_width) != len(self):
            raise ValueError(
                f"width must be a float or a sequence of length {len(self)}"
            )
        for line, w in zip(self, _width):
            line.width = w

    @property
    def color(self) -> NDArray[np.float32]:
        """2D array of RGBA color values."""
        return np.array([line.color for line in self], dtype=np.float32)

    @color.setter
    def color(self, color: str | Sequence[str]):
        col = as_color_array(color, len(self))
        for line, c in zip(self, col):
            line.color = c

    @property
    def style(self) -> Sequence[LineStyle]:
        """Array of line styles."""
        return np.array([line.style for line in self], dtype=np.float32)

    @style.setter
    def style(self, style: str | Sequence[str]):
        styles = as_any_1d_array(style, len(self))
        for line, s in zip(self, styles):
            line.style = s

    def with_hover_texts(self, text: str | Iterable[Any]) -> Self:
        if isinstance(text, str):
            texts = [text] * len(self)
        else:
            texts = [str(t) for t in text]
            if len(texts) != len(self):
                raise ValueError("Length of texts must match the number of lines.")
        for line, txt in zip(self, texts):
            line.with_hover_text(txt)
        return self

    def with_hover_template(
        self,
        template: str,
        extra: Any | None = None,
    ) -> Self:
        """Define hover template to the layer."""
        if self._backend_name in ("plotly", "bokeh"):  # conversion for HTML
            template = template.replace("\n", "<br>")
        params = parse_texts(template, len(self), extra)
        # set default format keys
        if "i" not in params:
            params["i"] = np.arange(len(self))
        texts = [
            template.format(**{k: v[i] for k, v in params.items()})
            for i in range(len(self))
        ]
        return self.with_hover_texts(texts)

    def _prep_markers(
        self,
        *,
        symbol: Symbol,
        size: float | None = None,
        alpha: float = 1.0,
        hatch: str | Hatch = Hatch.SOLID,
    ):
        from whitecanvas.layers.group import MarkerCollection

        markers = []
        size = theme._default("markers.size", size)
        for layer in self:
            color = layer.color
            mk = Markers(
                *layer.data, symbol=symbol, size=size, color=color, alpha=alpha,
                hatch=hatch, name=f"markers-of-{layer.name}", backend=self._backend_name
            )  # fmt: skip
            markers.append(mk)
        mcol = MarkerCollection(markers, name=f"markers-of-{self.name}")
        return mcol



"""
Plot API for whitecanvas.

>>> import numpy as np
>>> from whitecanvas import plot as plt
>>> plt.line(np.sin([1, 2, 3])).with_markers()
>>> plt.show()
"""

from whitecanvas.plot._canvases import (
    current_canvas,
    current_grid,
    figure,
    show,
    subplots,
    use,
)
from whitecanvas.plot._methods import (
    band,
    bars,
    cat,
    cat_x,
    cat_xy,
    cat_y,
    errorbars,
    hist,
    hline,
    infcurve,
    infline,
    kde,
    legend,
    line,
    markers,
    rug,
    spans,
    step,
    text,
    title,
    update_axes,
    update_font,
    update_labels,
    vline,
    xlabel,
    xlim,
    xticks,
    ylabel,
    ylim,
    yticks,
)

__all__ = [
    "line",
    "hline",
    "vline",
    "markers",
    "bars",
    "band",
    "hist",
    "spans",
    "step",
    "infcurve",
    "infline",
    "errorbars",
    "kde",
    "rug",
    "text",
    "cat",
    "cat_x",
    "cat_y",
    "cat_xy",
    "update_axes",
    "update_labels",
    "update_font",
    "legend",
    "figure",
    "show",
    "subplots",
    "current_grid",
    "current_canvas",
    "use",
    "xlim",
    "ylim",
    "xlabel",
    "ylabel",
    "xticks",
    "yticks",
    "title",
]



from __future__ import annotations

import weakref
from typing import TYPE_CHECKING

import numpy as np
from vispy import scene
from vispy.scene.visuals import Compound, Line, Text

if TYPE_CHECKING:
    from whitecanvas.backend.vispy.components3d.canvas import Camera, Canvas3D


class Axis3D(Compound):
    def __init__(self, canvas: Canvas3D, dim: int):
        name = "xyz"[dim]
        self._dim = dim
        self._line = Line(method="gl", width=2)
        self._text = Text(
            text=name,
            font_size=10,
            anchor_x="center",
            anchor_y="center",
        )
        self._canvas_ref = weakref.ref(canvas)
        super().__init__([self._line, self._text])
        self._update_pos([0, 0, 0], 1)
        self.parent = canvas._viewbox.scene

    def _plt_viewbox(self) -> scene.ViewBox:
        return self._canvas_ref()._viewbox

    def _plt_camera(self) -> Camera:
        return self._plt_viewbox().camera

    def _update_pos(
        self,
        origin: tuple[float, float, float],
        length: float | None = None,
    ):
        origin = np.asarray(origin)
        dr = np.zeros(3)
        if length is None:
            length = self._length()
        dr[self._dim] = length
        pos = np.stack([origin, origin + dr], axis=0)
        self._line.set_data(pos)
        self._text.pos = origin + dr * 1.1

    def _plt_get_limits(self) -> tuple[float, float]:
        start = self._line.pos[0]
        end = self._line.pos[1]
        return start[self._dim], end[self._dim]

    def _length(self) -> float:
        start = self._line.pos[0]
        end = self._line.pos[1]
        return np.sqrt(np.sum((end - start) ** 2))

    def _plt_set_limits(self, limits: tuple[float, float]):
        new_origin = self._line.pos[0]
        new_origin[self._dim] = limits[0]
        self._update_pos(new_origin, limits[1] - limits[0])
        canvas = self._canvas_ref()
        canvas._xaxis._update_pos(new_origin)
        canvas._yaxis._update_pos(new_origin)
        canvas._zaxis._update_pos(new_origin)

        start = self._line.pos[0]
        vecx = canvas._xaxis._line.pos[1] - start
        vecy = canvas._yaxis._line.pos[1] - start
        vecz = canvas._zaxis._line.pos[1] - start
        sizex = np.sqrt(np.dot(vecx, vecx))
        sizey = np.sqrt(np.dot(vecy, vecy))
        sizez = np.sqrt(np.dot(vecz, vecz))
        center = start + (vecx + vecy + vecz) / 2
        canvas._viewbox.camera.center = center
        canvas._viewbox.camera.scale_factor = max(sizex, sizey, sizez) * 1.5

    def _plt_get_color(self):
        return self._line.color

    def _plt_set_color(self, color):
        self._line.color = color

    def _plt_flip(self) -> None:
        camera = self._plt_camera()
        flipped = list(camera.flip)
        axis = 1 - self._dim
        flipped[axis] = not flipped[axis]
        camera.flip = tuple(flipped)


class AxisLabel3D:
    def __init__(self, axis: Axis3D):
        self._axis = axis

    def _plt_get_visible(self) -> bool:
        return self._axis._text.visible

    def _plt_set_visible(self, visible: bool):
        self._axis._text.visible = visible

    def _plt_get_text(self) -> str:
        return self._axis._text.text

    def _plt_set_text(self, text: str):
        self._axis._text.text = text

    def _plt_get_color(self):
        return np.array(self._axis._text.color).ravel()

    def _plt_set_color(self, color):
        self._axis._text.color = color

    def _plt_get_size(self) -> int:
        return self._axis._text.font_size

    def _plt_set_size(self, size: int):
        self._axis._text.font_size = size

    def _plt_get_fontfamily(self) -> str:
        return self._axis._text.face

    def _plt_set_fontfamily(self, family: str):
        self._axis._text.face = family



from __future__ import annotations

import numpy as np
from numpy.typing import NDArray
from plotly import graph_objects as go

from whitecanvas.backend import _not_implemented
from whitecanvas.backend.plotly._base import PlotlyHoverableLayer
from whitecanvas.utils.normalize import arr_color, as_color_array, rgba_str_color


class Mesh3D(PlotlyHoverableLayer[go.Mesh3d]):
    def __init__(self, verts: NDArray[np.floating], faces: NDArray[np.intp]):
        nfaces = faces.shape[0]
        self._props = {
            "x": verts[:, 0],
            "y": verts[:, 1],
            "z": verts[:, 2],
            "i": faces[:, 0],
            "j": faces[:, 1],
            "k": faces[:, 2],
            "type": "mesh3d",
            "showlegend": False,
            "visible": True,
            "facecolor": ["blue"] * nfaces,
            "vertexcolor": (0, 0, 255, 255),
        }
        self._verts = verts
        self._faces = faces
        PlotlyHoverableLayer.__init__(self)

    def _plt_get_data(self):
        return self._verts, self._faces

    def _plt_set_data(self, verts, faces):
        self._verts = verts
        self._faces = faces
        self._props["x"] = verts[:, 0]
        self._props["y"] = verts[:, 1]
        self._props["z"] = verts[:, 2]
        self._props["i"] = faces[:, 0]
        self._props["j"] = faces[:, 1]
        self._props["k"] = faces[:, 2]

    def _plt_get_ndata(self) -> int:
        return self._faces.shape[0]

    def _plt_get_face_color(self) -> NDArray[np.float32]:
        color = self._props["facecolor"]
        if len(color) == 0:
            return np.empty((0, 4), dtype=np.float32)
        return np.stack([arr_color(c) for c in color], axis=0)

    def _plt_set_face_color(self, color: NDArray[np.float32]):
        color = as_color_array(color, self._faces.shape[0])
        self._props["facecolor"] = [rgba_str_color(c) for c in color]

    _plt_get_face_hatch, _plt_set_face_hatch = _not_implemented.face_hatches()
    _plt_get_edge_color, _plt_set_edge_color = _not_implemented.edge_color()
    _plt_get_edge_width, _plt_set_edge_width = _not_implemented.edge_width()
    _plt_get_edge_style, _plt_set_edge_style = _not_implemented.edge_style()



import pandas as pd
from whitecanvas import new_jointgrid

def main():
    url = "https://raw.githubusercontent.com/mwaskom/seaborn-data/master/penguins.csv"
    df = pd.read_csv(url).dropna()

    joint = (
        new_jointgrid("matplotlib:qt")
        .with_hist_x(shape="step")  # show histogram as the x-marginal distribution
        .with_kde_y(width=2)  # show kde as the y-marginal distribution
        .with_rug(width=1)  # show rug plot for both marginal distributions
    )

    layer = (
        joint.cat(df, x="bill_length_mm", y="flipper_length_mm")
        .add_markers(color="species")
    )

    joint.add_legend()
    joint.show(block=True)

if __name__ == "__main__":
    main()



from __future__ import annotations

import numpy as np
from matplotlib.collections import PolyCollection

from whitecanvas.backend.matplotlib._base import MplMouseEventsMixin
from whitecanvas.protocols import BandProtocol, check_protocol
from whitecanvas.types import Hatch, LineStyle, Orientation


@check_protocol(BandProtocol)
class Band(PolyCollection, MplMouseEventsMixin):
    def __init__(
        self,
        t: np.ndarray,
        ydata0: np.ndarray,
        ydata1: np.ndarray,
        orient: Orientation,
    ):
        if orient.is_vertical:
            fw = np.stack([t, ydata0], axis=1)
            bw = np.stack([t[::-1], ydata1[::-1]], axis=1)
        else:
            fw = np.stack([ydata0, t], axis=1)
            bw = np.stack([ydata1[::-1], t[::-1]], axis=1)
        verts = np.concatenate([fw, bw], axis=0)
        self._edge_style = LineStyle.SOLID
        super().__init__([verts], closed=True, picker=True)
        self.set_edgecolor("#00000000")
        self._t = t
        self._y0 = ydata0
        self._y1 = ydata1
        MplMouseEventsMixin.__init__(self)

    ##### XYYDataProtocol #####
    def _plt_get_vertical_data(self):
        return self._t, self._y0, self._y1

    def _plt_get_horizontal_data(self):
        return self._t, self._y0, self._y1

    def _plt_set_vertical_data(self, t, ydata0, ydata1):
        verts = np.concatenate(
            [
                np.stack([t, ydata0], axis=1),
                np.stack([t[::-1], ydata1[::-1]], axis=1),
            ],
            axis=0,
        )
        self.set_verts([verts])
        self._t = t
        self._y0 = ydata0
        self._y1 = ydata1

    def _plt_set_horizontal_data(self, t, ydata0, ydata1):
        verts = np.concatenate(
            [
                np.stack([ydata0, t], axis=1),
                np.stack([ydata1[::-1], t[::-1]], axis=1),
            ],
            axis=0,
        )
        self.set_verts([verts])
        self._t = t
        self._y0 = ydata0
        self._y1 = ydata1

    def _plt_get_face_color(self):
        return self.get_facecolor()[0]

    def _plt_set_face_color(self, color):
        self.set_facecolor(color)

    def _plt_get_face_hatch(self) -> Hatch:
        return Hatch(self.get_hatch() or "")

    def _plt_set_face_hatch(self, pattern: Hatch):
        if pattern is Hatch.SOLID:
            ptn = None
        else:
            ptn = pattern.value
        self.set_hatch(ptn)

    def _plt_get_edge_color(self):
        return self.get_edgecolor()[0]

    def _plt_set_edge_color(self, color):
        self.set_edgecolor(color)

    def _plt_get_edge_width(self):
        return self.get_linewidth()[0]

    def _plt_set_edge_width(self, width: float):
        self.set_linewidth(width)

    def _plt_get_edge_style(self):
        return self._edge_style

    def _plt_set_edge_style(self, style: LineStyle):
        self.set_linestyle(style.value)
        self._edge_style = style

    def _plt_get_antialias(self):
        return self.get_antialiased()

    def _plt_set_antialias(self, antialias: bool):
        self.set_antialiased(antialias)

    def _on_hover(self, event=None):
        if self._hover_texts is None or not self._plt_get_visible():
            return
        contains, ind = self.contains(event)
        if not contains:
            return
        return self._hover_texts[0]

    def _plt_set_hover_text(self, text: str):
        self._hover_texts = [text]



from __future__ import annotations

import numpy as np
from matplotlib.collections import LineCollection
from matplotlib.lines import Line2D
from numpy.typing import NDArray

from whitecanvas.backend.matplotlib._base import MplMouseEventsMixin
from whitecanvas.protocols import LineProtocol, MultiLineProtocol, check_protocol
from whitecanvas.types import LineStyle
from whitecanvas.utils.normalize import as_color_array
from whitecanvas.utils.type_check import is_real_number


@check_protocol(LineProtocol)
class MonoLine(Line2D, MplMouseEventsMixin):
    def __init__(self, xdata, ydata):
        super().__init__(
            xdata,
            ydata,
            linewidth=1,
            linestyle="-",
            color="blue",
            markersize=0,
            picker=True,
            pickradius=5,
        )
        MplMouseEventsMixin.__init__(self)

    ##### XYDataProtocol #####
    def _plt_get_data(self):
        return self.get_data()

    def _plt_set_data(self, xdata, ydata):
        self.set_data(xdata, ydata)

    ##### HasEdges #####
    def _plt_get_edge_width(self) -> float:
        return self.get_linewidth()

    def _plt_set_edge_width(self, width: float):
        self.set_linewidth(width)

    def _plt_get_edge_style(self) -> LineStyle:
        return LineStyle(self.get_linestyle())

    def _plt_set_edge_style(self, style: LineStyle):
        self.set_linestyle(style.value)

    def _plt_get_edge_color(self) -> NDArray[np.float32]:
        return self.get_color()

    def _plt_set_edge_color(self, color: NDArray[np.float32]):
        self.set_color(color)

    def _plt_get_antialias(self) -> bool:
        return self.get_antialiased()

    def _plt_set_antialias(self, antialias: bool):
        self.set_antialiased(antialias)


@check_protocol(MultiLineProtocol)
class MultiLine(LineCollection, MplMouseEventsMixin):
    def __init__(self, data: list[NDArray[np.floating]]):
        # data: list of (N, 2)
        super().__init__(data, linewidths=1, picker=True, pickradius=5)
        self._ndata = len(data)
        self._linestyle = [LineStyle.SOLID] * self._ndata
        MplMouseEventsMixin.__init__(self)

    ##### XYDataProtocol #####
    def _plt_get_data(self):
        return self.get_segments()

    def _plt_set_data(self, data: list[NDArray[np.floating]]):
        self.set_segments(data)
        self._ndata = len(data)

    ##### HasEdges #####
    def _plt_get_edge_width(self) -> NDArray[np.floating]:
        return self.get_linewidth()

    def _plt_set_edge_width(self, width: float | NDArray[np.floating]):
        if is_real_number(width):
            width = np.full(self._ndata, width)
        else:
            width = np.asarray(width)
        self.set_linewidth(width)

    def _plt_get_edge_style(self) -> list[LineStyle]:
        return self._linestyle

    def _plt_set_edge_style(self, style: LineStyle | list[LineStyle]):
        if isinstance(style, LineStyle):
            _style = [style.value] * self._ndata
            _style_enum = [style] * self._ndata
        else:
            _style = [s.value for s in style]
            _style_enum = style
        self.set_linestyle(_style)
        self._linestyle = _style_enum

    def _plt_get_edge_color(self) -> NDArray[np.float32]:
        return self.get_color()

    def _plt_set_edge_color(self, color: NDArray[np.float32]):
        self.set_color(as_color_array(color, self._ndata))

    def _plt_get_antialias(self) -> bool:
        return self.get_antialiased()[0]

    def _plt_set_antialias(self, antialias: bool):
        self.set_antialiased(antialias)



import numpy as np
from whitecanvas import new_canvas

def main():
    canvas = new_canvas("pyqtgraph:qt")
    rng = np.random.default_rng(14872)
    data = rng.poisson(4, (8, 10))
    heatmap = canvas.add_heatmap(data, cmap="inferno").with_text()
    canvas.show(block=True)

if __name__ == "__main__":
    main()



from __future__ import annotations

from typing import Sequence

import numpy as np
from numpy.typing import ArrayLike, NDArray

from whitecanvas.utils.normalize import as_array_1d


def check_array_input(
    x: Sequence[float],
    data: Sequence[ArrayLike],
) -> tuple[NDArray[np.floating], list[NDArray[np.floating]], list[str]]:
    x = as_array_1d(x)
    if len(x) != len(data):
        raise ValueError(f"len(x) != len(data), got {len(x)} and {len(data)}")
    return x, data



from __future__ import annotations

from whitecanvas.types import XYData


def norm_label_text(strings: str | list[str], data: XYData) -> list[str]:
    if not isinstance(strings, str):
        strings = list(strings)
        if len(strings) != data.x.size:
            raise ValueError(
                f"Length of strings ({len(strings)}) does not match data size "
                f"({data.x.size})"
            )
        return strings
    if "{x" in strings or "{y" in strings or "{i" in strings:
        try:
            strings.format(x=0.0, y=0.0, i=0)  # dry run
        except Exception:
            strings = [strings] * data.x.size
        else:
            strings = [
                strings.format(x=x, y=y, i=i) for i, (x, y) in enumerate(zip(*data))
            ]
    else:
        strings = [strings] * data.x.size
    return strings



from functools import singledispatch

import bokeh.models as bk_models

from whitecanvas.backend.bokeh._base import (
    to_bokeh_hatch,
    to_bokeh_line_style,
    to_bokeh_symbol,
)
from whitecanvas.layers import _legend as _leg
from whitecanvas.utils.normalize import hex_color

_DATA = bk_models.ColumnDataSource({"x": [0], "y": [0]})


@singledispatch
def make_sample_item(item) -> "list[bk_models.GlyphRenderer] | None":
    return None


@make_sample_item.register
def _(item: _leg.LineLegendItem):
    line = bk_models.Line(
        x="x", y="x", line_dash=to_bokeh_line_style(item.style),
        line_color=hex_color(item.color), line_width=item.width,
    )  # fmt: skip
    return [bk_models.GlyphRenderer(glyph=line, data_source=_DATA, visible=False)]


@make_sample_item.register
def _(item: _leg.MarkersLegendItem):
    marker, angle = to_bokeh_symbol(item.symbol)
    sc = bk_models.Scatter(
        x="x", y="x", fill_color=hex_color(item.face.color),
        marker=marker, angle=angle,
        hatch_pattern=to_bokeh_hatch(item.face.hatch),
        line_color=hex_color(item.edge.color),
        line_width=item.edge.width, line_dash=to_bokeh_line_style(item.edge.style),
        size=item.size,
    )  # fmt: skip
    return [bk_models.GlyphRenderer(glyph=sc, data_source=_DATA, visible=False)]


@make_sample_item.register
def _(item: _leg.BarLegendItem):
    quad = bk_models.Quad(
        left="x",
        right="x",
        bottom="x",
        top="x",
        fill_color=hex_color(item.face.color),
        hatch_pattern=to_bokeh_hatch(item.face.hatch),
        line_color=hex_color(item.edge.color),
        line_width=item.edge.width,
        line_dash=to_bokeh_line_style(item.edge.style),
    )
    return [bk_models.GlyphRenderer(glyph=quad, data_source=_DATA, visible=False)]


@make_sample_item.register
def _(item: _leg.PlotLegendItem):
    return make_sample_item(item.line) + make_sample_item(item.markers)


@make_sample_item.register
def _(item: _leg.ErrorLegendItem):
    line = bk_models.Line(
        x="x", y="x", line_dash=to_bokeh_line_style(item.style),
        line_color=hex_color(item.color), line_width=item.width,
    )  # fmt: skip
    return [bk_models.GlyphRenderer(glyph=line, data_source=_DATA, visible=False)]


@make_sample_item.register
def _(item: _leg.LineErrorLegendItem):
    items = make_sample_item(item.line)
    if item.xerr is not None:
        items.extend(make_sample_item(item.xerr))
    if item.yerr is not None:
        items.extend(make_sample_item(item.yerr))
    return items


@make_sample_item.register
def _(item: _leg.MarkerErrorLegendItem):
    items = make_sample_item(item.markers)
    if item.xerr is not None:
        items.extend(make_sample_item(item.xerr))
    if item.yerr is not None:
        items.extend(make_sample_item(item.yerr))
    return items


@make_sample_item.register
def _(item: _leg.PlotErrorLegendItem):
    items = make_sample_item(item.plot)
    if item.xerr is not None:
        items.extend(make_sample_item(item.xerr))
    if item.yerr is not None:
        items.extend(make_sample_item(item.yerr))
    return items


@make_sample_item.register
def _(item: _leg.StemLegendItem):
    return make_sample_item(item.markers)


@make_sample_item.register
def _(item: _leg.TitleItem):
    line = bk_models.Line(x="x", y="x", line_color="#00000000", line_width=0)
    return [bk_models.GlyphRenderer(glyph=line, data_source=_DATA, visible=False)]



from whitecanvas.canvas.dataframe._both_cat import XYCatPlotter
from whitecanvas.canvas.dataframe._feature_cat import CatPlotter
from whitecanvas.canvas.dataframe._joint_cat import JointCatPlotter
from whitecanvas.canvas.dataframe._one_cat import XCatPlotter, YCatPlotter

__all__ = [
    "CatPlotter",
    "XCatPlotter",
    "YCatPlotter",
    "XYCatPlotter",
    "JointCatPlotter",
]



from __future__ import annotations

import inspect
import math
from typing import TYPE_CHECKING, Any, Callable, Generic

import numpy as np
from typing_extensions import Concatenate, ParamSpec

from whitecanvas.backend import Backend
from whitecanvas.layers._primitive.line import LineMixin
from whitecanvas.protocols import LineProtocol
from whitecanvas.types import ColorType, LineStyle, Rect

if TYPE_CHECKING:
    from typing_extensions import Self

    from whitecanvas.canvas import Canvas

_P = ParamSpec("_P")


class InfCurve(LineMixin[LineProtocol], Generic[_P]):
    _backend_class_name = "MonoLine"
    _NO_PADDING_NEEDED = True

    def __init__(
        self,
        model: Callable[[Concatenate[np.ndarray, _P]], np.ndarray],
        *,
        bounds: tuple[float, float] = (-np.inf, np.inf),
        name: str | None = None,
        color: ColorType = "blue",
        alpha: float = 1,
        width: float = 1,
        style: LineStyle | str = LineStyle.SOLID,
        antialias: bool = True,
        backend: Backend | str | None = None,
    ):
        _lower, _upper = bounds
        if not np.isfinite(_lower):
            _lower = 0
        if not np.isfinite(_upper):
            _upper = 1
        super().__init__(name=name)

        xdata = np.array([_lower, _upper])
        ydata, self._y_hint, self._params_ready = _try_init_ydata(model, xdata)
        self._backend = self._create_backend(Backend(backend), xdata, ydata)
        self.update(
            color=color, width=width, style=style, alpha=alpha,
            antialias=antialias
        )  # fmt: skip

        self._bounds = bounds
        self._model = model
        self._args = ()
        self._kwargs = {}
        self._linspace_num = 256

    def update_params(self, *args: _P.args, **kwargs: _P.kwargs) -> Self:
        """Set the parameters of the model function."""
        xdata, _ = self._backend._plt_get_data()
        ydata = self._model(xdata, *args, **kwargs)
        self._backend._plt_set_data(xdata, ydata)
        self._args, self._kwargs = args, kwargs
        if not self._params_ready:
            self._params_ready = True
            self._canvas().x.events.lim.emit(self._canvas().x.lim)
        return self

    @property
    def params(self) -> tuple[tuple, dict[str, Any]]:
        """The parameters of the model function."""
        return self._args, self._kwargs

    @property
    def model(self) -> Callable[[Concatenate[np.ndarray, _P]], np.ndarray]:
        """The model function of the layer."""
        return self._model

    def with_hover_text(self, text: str) -> Self:
        if not isinstance(text, str):
            raise TypeError(f"Hover text must be str, got {type(text)}.")
        self._backend._plt_set_hover_text([text] * self._linspace_num)
        return self

    def _connect_canvas(self, canvas: Canvas):
        canvas.x.events.lim.connect(self._recalculate_line, max_args=1)
        self._recalculate_line(canvas.x.lim)
        super()._connect_canvas(canvas)

    def _disconnect_canvas(self, canvas: Canvas):
        canvas.x.events.lim.disconnect(self._recalculate_line)
        super()._disconnect_canvas(canvas)

    def _recalculate_line(self, lim: tuple[float, float]) -> None:
        x0, x1 = lim
        b0, b1 = self._bounds
        x0 = max(x0, b0)
        x1 = min(x1, b1)
        if x0 >= x1:
            xdata, ydata = np.array([]), np.array([])
        else:
            xdata = np.linspace(x0, x1, self._linspace_num)
            if self._params_ready:
                ydata = self._model(xdata, *self._args, **self._kwargs)
            else:
                return
        self._backend._plt_set_data(xdata, ydata)


def _try_init_ydata(
    model: Callable[[Concatenate[np.ndarray, _P]], np.ndarray],
    xdata: np.ndarray,
):
    try:
        _sig = inspect.signature(model)
    except ValueError:
        return _try_init_ydata_for_ufunc(model, xdata)
    try:
        _sig.bind(xdata)
    except TypeError:
        ydata = np.zeros_like(xdata)
        yhint = None
        ready = False
    else:
        ydata = model(xdata)
        yhint = ydata.min(), ydata.max()
        ready = True
    return ydata, yhint, ready


def _try_init_ydata_for_ufunc(
    model: Callable[[Concatenate[np.ndarray, _P]], np.ndarray],
    xdata: np.ndarray,
):
    try:
        ydata = model(xdata)
    except TypeError:
        ydata = np.zeros_like(xdata)
        yhint = None
        ready = False
    else:
        yhint = ydata.min(), ydata.max()
        ready = True
    return ydata, yhint, ready


class InfLine(LineMixin[LineProtocol]):
    _backend_class_name = "MonoLine"
    _NO_PADDING_NEEDED = True

    def __init__(
        self,
        pos: tuple[float, float] = (0, 0),
        angle: float = 0.0,
        *,
        name: str | None = None,
        color: ColorType = "blue",
        alpha: float = 1,
        width: float = 1,
        style: LineStyle | str = LineStyle.SOLID,
        antialias: bool = True,
        backend: Backend | str | None = None,
    ):
        self._pos = pos
        self._is_vline = angle % 180 == 90
        if self._is_vline:
            self._tan = 0  # not used
            self._intercept = self._pos[0]
        else:
            _radian = math.radians(angle) % (2 * math.pi)
            self._tan = math.tan(_radian)
            self._intercept = self._pos[1] - self._tan * self._pos[0]
        super().__init__(name=name)
        self._backend = self._create_backend(Backend(backend), np.zeros(1), np.zeros(1))
        self.update(
            color=color, width=width, style=style, alpha=alpha,
            antialias=antialias,
        )  # fmt: skip
        self._last_rect = Rect(0, 0, 0, 0)

    @property
    def pos(self) -> tuple[float, float]:
        """One of the points on the line."""
        return self._pos

    @pos.setter
    def pos(self, pos: tuple[float, float]):
        self._pos = pos
        self._recalculate_line(self._last_rect)

    @property
    def angle(self) -> float:
        """The angle of the line in degrees."""
        if self._is_vline:
            return 90.0
        else:
            return math.degrees(math.atan(self._tan))

    @angle.setter
    def angle(self, angle: float):
        self._is_vline = angle % 180 == 90
        if self._is_vline:
            self._tan = 0  # not used
            self._intercept = self._pos[0]
        else:
            _radian = math.radians(angle) % (2 * math.pi)
            self._tan = math.tan(_radian)
            self._intercept = self._pos[1] - self._tan * self._pos[0]
        self._recalculate_line(self._last_rect)

    def with_hover_text(self, text: str) -> Self:
        if not isinstance(text, str):
            raise TypeError(f"Hover text must be str, got {type(text)}.")
        self._backend._plt_set_hover_text([text] * 2)
        return self

    def _connect_canvas(self, canvas: Canvas):
        canvas.events.lims.connect(self._recalculate_line, max_args=1)
        self._recalculate_line(canvas.lims)
        self._last_rect = canvas.lims
        super()._connect_canvas(canvas)

    def _disconnect_canvas(self, canvas: Canvas):
        canvas.events.lims.disconnect(self._recalculate_line)
        super()._disconnect_canvas(canvas)

    def _recalculate_line(self, rect: Rect) -> None:
        x0, x1, y0, y1 = rect
        self._last_rect = rect
        if self._is_vline:
            x = self._intercept
            xdata = np.array([x, x])
            ydata = np.array([y0, y1])
        else:
            xdata = np.array([x0, x1])
            ydata = self._tan * xdata + self._intercept
        self._backend._plt_set_data(xdata, ydata)



from __future__ import annotations

import pyqtgraph as pg
from qtpy import QtCore, QtGui
from qtpy import QtWidgets as QtW

from whitecanvas.types import Rect


class PyQtLayer:
    def _plt_get_visible(self) -> bool:
        return self.isVisible()

    def _plt_set_visible(self, visible: bool):
        self.setVisible(visible)

    def _plt_set_zorder(self, zorder: int):
        self.setZValue(zorder)


class InsetPlotItem(pg.GraphicsWidget):
    def __init__(self, item: pg.PlotItem, rect: Rect):
        pg.GraphicsWidget.__init__(self)
        self._layout = QtW.QGraphicsGridLayout()
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(0)
        self.setLayout(self._layout)
        self._layout.addItem(item, 0, 0)
        self._item = item
        self._rect = rect

    def mouseDragEvent(self, ev):
        self._item.vb.mouseDragEvent(ev)

    def setParentItem(self, parent: pg.ViewBox):
        super().setParentItem(parent)
        parent.geometryChanged.connect(self._update_geometry)
        parent.sigResized.connect(self._update_geometry)

    def _update_geometry(self, *_):
        vb_rect: QtCore.QRect = self.parentItem().boundingRect()
        width = vb_rect.width()
        height = vb_rect.height()
        self.setGeometry(
            int(width * self._rect.left),
            int(height * (1 - self._rect.top)),
            int(width * self._rect.width),
            int(height * self._rect.height),
        )


class PyQtAxis(pg.AxisItem):
    def __init__(self, orientation, **kwargs):
        super().__init__(orientation, **kwargs)
        self.style["tickRotation"] = 0.0

    def drawPicture(self, p: QtGui.QPainter, axisSpec, tickSpecs, textSpecs):
        p.setRenderHint(p.RenderHint.Antialiasing, False)
        p.setRenderHint(p.RenderHint.TextAntialiasing, True)

        ## draw long line along axis
        pen, p1, p2 = axisSpec
        p.setPen(pen)
        p.drawLine(p1, p2)
        # p.translate(0.5,0)  ## resolves some damn pixel ambiguity

        ## draw ticks
        for pen, p1, p2 in tickSpecs:
            p.setPen(pen)
            p.drawLine(p1, p2)

        # Draw all text
        if self.style["tickFont"] is not None:
            p.setFont(self.style["tickFont"])
        p.setPen(self.textPen())
        bounding = self.boundingRect().toAlignedRect()
        p.setClipRect(bounding)
        rot = self.style["tickRotation"]
        if abs(rot) < 1e-3:
            for rect, flags, text in textSpecs:
                p.drawText(rect, int(flags), text)
        else:
            for rect, flags, text in textSpecs:
                rect: QtCore.QRect
                p.save()
                p.translate(rect.center())
                p.rotate(rot)
                p.drawText(rect.translated(-rect.center()), int(flags), text)
                p.rotate(-rot)
                p.restore()

    def tickRotation(self) -> float:
        return -self.style["tickRotation"]

    def setTickRotation(self, rotation: float) -> None:
        self.style["tickRotation"] = -rotation
        self.picture = None
        self.update()



from __future__ import annotations

import numpy as np
from cmap import Colormap
from numpy.typing import NDArray

from whitecanvas import protocols
from whitecanvas.backend.mock._base import (
    BaseMockLayer,
    MockHasData,
    MockHasEdges,
    MockHasFaces,
    MockHasMouseEvents,
    MockHasMultiEdges,
    MockHasMultiFaces,
)
from whitecanvas.types import Alignment, Orientation, Symbol
from whitecanvas.utils.normalize import as_color_array
from whitecanvas.utils.type_check import is_real_number


@protocols.check_protocol(protocols.BandProtocol)
class Band(MockHasData, MockHasFaces, MockHasEdges, MockHasMouseEvents):
    def __init__(
        self,
        t: np.ndarray,
        ydata0: np.ndarray,
        ydata1: np.ndarray,
        orient: Orientation,
    ):
        super().__init__((t, ydata0, ydata1))

    ##### XYYDataProtocol #####
    def _plt_get_vertical_data(self):
        return self._plt_get_data()

    def _plt_get_horizontal_data(self):
        return self._plt_get_data()

    def _plt_set_vertical_data(self, t, ydata0, ydata1):
        self._plt_set_data(t, ydata0, ydata1)

    def _plt_set_horizontal_data(self, t, ydata0, ydata1):
        self._plt_set_data(t, ydata0, ydata1)


@protocols.check_protocol(protocols.BarProtocol)
class Bars(MockHasData, MockHasMultiFaces, MockHasMultiEdges, MockHasMouseEvents):
    def __init__(self, xlow, xhigh, ylow, yhigh):
        super().__init__((xlow, xhigh, ylow, yhigh))

    def _plt_get_ndata(self) -> int:
        return self._plt_get_data()[0].size


@protocols.check_protocol(protocols.ImageProtocol)
class Image(BaseMockLayer):
    def __init__(self, data: np.ndarray):
        super().__init__()
        self._data = data
        self._clim = (0, 1)
        self._cmap = Colormap("gray")
        self._scale = (1, 1)
        self._translation = (0, 0)

    def _plt_get_data(self) -> np.ndarray:
        return self._data

    def _plt_set_data(self, data: np.ndarray):
        self._data = data

    def _plt_get_colormap(self) -> Colormap:
        return self._cmap

    def _plt_set_colormap(self, cmap: Colormap):
        self._cmap = cmap

    def _plt_get_clim(self) -> tuple[float, float]:
        return self._clim

    def _plt_set_clim(self, clim: tuple[float, float]):
        self._clim = clim

    def _plt_get_scale(self) -> tuple[float, float]:
        return self._scale

    def _plt_set_scale(self, scale: tuple[float, float]):
        self._scale = scale

    def _plt_get_translation(self) -> tuple[float, float]:
        return self._translation

    def _plt_set_translation(self, translation: tuple[float, float]):
        self._translation = translation


@protocols.check_protocol(protocols.LineProtocol)
class MonoLine(MockHasData, MockHasEdges, MockHasMouseEvents):
    def __init__(self, xdata, ydata):
        super().__init__((xdata, ydata))
        self._antialias = True

    def _plt_get_antialias(self) -> bool:
        return self._antialias

    def _plt_set_antialias(self, antialias: bool):
        self._antialias = antialias


@protocols.check_protocol(protocols.MultiLineProtocol)
class MultiLine(BaseMockLayer, MockHasMultiEdges, MockHasMouseEvents):
    def __init__(self, data: list[np.ndarray]):
        super().__init__()
        self._data = data
        self._antialias = True

    def _plt_get_data(self) -> list[np.ndarray]:
        return self._data

    def _plt_set_data(self, data: list[np.ndarray]):
        self._data = data

    def _plt_get_antialias(self) -> bool:
        return self._antialias

    def _plt_set_antialias(self, antialias: bool):
        self._antialias = antialias

    def _plt_get_ndata(self) -> int:
        return len(self._plt_get_data())


@protocols.check_protocol(protocols.MarkersProtocol)
class Markers(MockHasData, MockHasMultiFaces, MockHasMultiEdges, MockHasMouseEvents):
    def __init__(self, xdata, ydata):
        super().__init__((xdata, ydata))
        self._symbol = Symbol.CIRCLE
        self._size = np.full(xdata.shape, 10.0, dtype=np.float32)

    def _plt_get_ndata(self) -> int:
        return self._plt_get_data()[0].size

    def _plt_get_symbol(self) -> Symbol:
        return self._symbol

    def _plt_set_symbol(self, symbol: Symbol):
        self._symbol = symbol

    def _plt_get_symbol_size(self):
        return self._size

    def _plt_set_symbol_size(self, size):
        if is_real_number(size):
            size = np.full(self._size.shape, size, dtype=np.float32)
        self._size = size


@protocols.check_protocol(protocols.TextProtocol)
class Texts(MockHasData, MockHasMultiFaces, MockHasMultiEdges):
    def __init__(self, x, y, text):
        super().__init__((x, y))
        self._text = text
        ndata = len(text)
        self._fontfamily = ["sans-serif"] * ndata
        self._fontsize = np.full(ndata, 10.0, dtype=np.float32)
        self._text_color = np.zeros((ndata, 4), dtype=np.float32)
        self._rotation = np.zeros(ndata, dtype=np.float32)
        self._anchor = [Alignment.LEFT] * ndata

    def _plt_get_ndata(self) -> int:
        return len(self._text)

    def _plt_get_text(self) -> list[str]:
        return self._text

    def _plt_set_text(self, text: list[str]):
        self._text = text

    def _plt_get_text_color(self):
        return self._text_color

    def _plt_set_text_color(self, color):
        self._text_color = as_color_array(color, self._plt_get_ndata())

    def _plt_get_text_size(self):
        return self._fontsize

    def _plt_set_text_size(self, size):
        if is_real_number(size):
            size = np.full(self._fontsize.shape, size, dtype=np.float32)
        self._fontsize = size

    def _plt_get_text_position(self):
        return self._plt_get_data()

    def _plt_set_text_position(self, position):
        self._plt_set_data(*position)

    def _plt_get_text_anchor(self) -> Alignment:
        return self._anchor

    def _plt_set_text_anchor(self, anc: Alignment):
        self._anchor = anc

    def _plt_get_text_rotation(self):
        return self._rotation

    def _plt_set_text_rotation(self, rotation):
        if is_real_number(rotation):
            rotation = np.full(self._rotation.shape, rotation, dtype=np.float32)
        self._rotation = rotation

    def _plt_get_text_fontfamily(self) -> str:
        return self._fontfamily

    def _plt_set_text_fontfamily(self, family: str):
        self._fontfamily = family


@protocols.check_protocol(protocols.VectorsProtocol)
class Vectors(MockHasData, MockHasEdges):
    def __init__(self, x0, dx, y0, dy):
        super().__init__((x0, dx, y0, dy))

    def _plt_get_ndata(self) -> int:
        return self._plt_get_data()[0].size

    def _plt_get_edge_color(self) -> NDArray[np.float32]:
        if not hasattr(self, "_edge_color"):
            ndata = self._plt_get_ndata()
            self._edge_color = np.zeros((ndata, 4), dtype=np.float32)
        return self._edge_color

    def _plt_set_edge_color(self, color: NDArray[np.float32]):
        self._edge_color = as_color_array(color, self._plt_get_ndata())

    def _plt_get_antialias(self) -> bool:
        return self._antialias

    def _plt_set_antialias(self, antialias: bool):
        self._antialias = antialias



from __future__ import annotations

import warnings
import weakref
from typing import TYPE_CHECKING

import numpy as np
from vispy.scene import visuals

from whitecanvas.backend.vispy.line import _make_connection, _safe_concat

if TYPE_CHECKING:
    from whitecanvas.backend.vispy.canvas import Canvas


class LineProperties:
    def __init__(self):
        self.visible = False
        self.color = np.ones(4, dtype=np.float32)
        self.width = 1.0
        self.style = "solid"


class GridLines(visuals.Compound):
    def __init__(self, canvas: Canvas):
        self._xscale = 0
        self._yscale = 0
        self._canvas_ref = weakref.ref(canvas)
        self._xlines = visuals.Line()
        self._ylines = visuals.Line()
        self._xprops = LineProperties()
        self._yprops = LineProperties()
        super().__init__([self._xlines, self._ylines])
        canvas._viewbox.add(self)
        self.order = -99999

    def set_x_grid_lines(self, visible: bool, color, width: float, style):
        self._xprops.visible = visible
        self._xprops.color = color
        self._xprops.width = width
        self._xprops.style = style
        self.update()

    def set_y_grid_lines(self, visible: bool, color, width: float, style):
        self._yprops.visible = visible
        self._yprops.color = color
        self._yprops.width = width
        self._yprops.style = style
        self.update()

    def _prepare_draw(self, view):
        if not (self._xprops.visible or self._yprops.visible):
            return super()._prepare_draw(view)
        if canvas := self._canvas_ref():
            rect = canvas._camera.rect
            xmin, xmax = rect.left, rect.right
            ymin, ymax = rect.bottom, rect.top
            if self._xprops.visible:
                xmajor_pos = canvas._xticks._get_ticker()._get_tick_frac_labels()[0]
                xmajor = xmajor_pos * (xmax - xmin) + xmin
                xdata = [np.array([[x, ymin], [x, ymax]]) for x in xmajor]
                self._xlines.set_data(
                    pos=_safe_concat(xdata),
                    color=self._xprops.color,
                    width=self._xprops.width,
                    connect=_make_connection(xdata),
                )
            if self._yprops.visible:
                ymajor_pos = canvas._yticks._get_ticker()._get_tick_frac_labels()[0]
                ymajor = ymajor_pos * (ymax - ymin) + ymin
                ydata = [np.array([[xmin, y], [xmax, y]]) for y in ymajor]
                self._ylines.set_data(
                    pos=_safe_concat(ydata),
                    color=self._yprops.color,
                    width=self._yprops.width,
                    connect=_make_connection(ydata),
                )
        return super()._prepare_draw(view)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TypeVar

import numpy as np
from numpy.typing import NDArray

from whitecanvas.layers.tabular._df_compat import DataFrameWrapper

_DF = TypeVar("_DF")


class JitterBase(ABC):
    @abstractmethod
    def map(self, src: DataFrameWrapper[_DF]) -> NDArray[np.floating]:
        """Map the source data to jittered data."""


class IdentityJitter(JitterBase):
    """No jittering."""

    def __init__(self, by: str):
        if not isinstance(by, str):
            raise TypeError(f"Only str is allowed, got {type(by)}")
        self._by = by

    def map(self, src: DataFrameWrapper[_DF]) -> NDArray[np.floating]:
        return src[self._by]

    def check(self, src: DataFrameWrapper[_DF]) -> IdentityJitter:
        if self._by not in src:
            raise ValueError(f"Column {self._by} not found in the data frame.")
        if src[self._by].dtype.kind not in "iufb":
            raise ValueError(f"Column {self._by} is not numeric.")
        return self


class CategoricalLikeJitter(JitterBase):
    def __init__(self, by: str | tuple[str, ...], mapping: dict[tuple, float]):
        self._by = _tuple(by)
        self._mapping = mapping

    def _map(self, src: DataFrameWrapper[_DF]) -> NDArray[np.floating]:
        # only map the categorical data to real numbers
        args = [src[b] for b in self._by]
        out = np.zeros(len(src), dtype=np.float32)
        for row, pos in self._mapping.items():
            arrs = [a == r for a, r in zip(args, row) if a.size > 0]
            if len(arrs) == 0:
                continue
            sl = np.all(np.column_stack(arrs), axis=1)
            out[sl] = pos
        return out

    @property
    def by(self) -> tuple[str, ...]:
        return self._by


class CategoricalJitter(CategoricalLikeJitter):
    """Jitter for categorical data."""

    def map(self, src: DataFrameWrapper[_DF]) -> NDArray[np.floating]:
        return self._map(src)


class UniformJitter(CategoricalLikeJitter):
    """Jitter with uniform distribution."""

    def __init__(
        self,
        by: str | tuple[str, ...],
        mapping: dict[tuple, float],
        extent: float = 0.8,
        seed: int | None = 0,
    ):
        super().__init__(by, mapping)
        self._rng = np.random.default_rng(seed)
        self._extent = extent

    def map(self, src: DataFrameWrapper[_DF]) -> NDArray[np.floating]:
        w = self._extent
        jitter = self._rng.uniform(-w / 2, w / 2, size=len(src))
        return self._map(src) + jitter


class SwarmJitter(CategoricalLikeJitter):
    """Jitter for swarm plot."""

    def __init__(
        self,
        by: str | tuple[str, ...],
        mapping: dict[tuple, float],
        value: str,
        limits: tuple[float, float],
        extent: float = 0.8,
    ):
        super().__init__(by, mapping)
        self._value = value
        self._extent = extent
        self._limits = limits

    def _get_bins(self, src: DataFrameWrapper[_DF]) -> int:
        return 25  # just for now

    def map(self, src: DataFrameWrapper[_DF]) -> NDArray[np.floating]:
        values = src[self._value]
        vmin, vmax = self._limits
        nbin = self._get_bins(src)
        dv = (vmax - vmin) / nbin
        # bin index that each value belongs to
        v_indices = np.floor((values - vmin) / dv).astype(np.int32)
        v_indices[v_indices == nbin] = nbin - 1

        args = [src[b] for b in self._by]
        offset_pre = np.zeros(len(src), dtype=np.float32)
        for row in self._mapping.keys():
            sl = np.all(np.column_stack([a == r for a, r in zip(args, row)]), axis=1)
            offset_pre[sl] = self._map_one(v_indices[sl], nbin)

        offset_max = np.abs(offset_pre).max()
        if offset_max == 0:  # when the swarm plot is sparse
            return self._map(src)
        width_default = dv * offset_max * 0.6
        offsets = offset_pre / offset_max * min(self._extent / 2, width_default)
        out = self._map(src) + offsets
        return out

    def _map_one(self, indices: NDArray[np.int32], nbin: int) -> NDArray[np.floating]:
        offset_count = np.zeros(nbin, dtype=np.int32)
        offset_pre = np.zeros_like(indices, dtype=np.int32)
        for i, idx in enumerate(indices):
            c = offset_count[idx]
            if c % 2 == 0:
                offset_pre[i] = c / 2
            else:
                offset_pre[i] = -(c + 1) / 2
            offset_count[idx] += 1
        return offset_pre


def _tuple(x) -> tuple[str, ...]:
    if isinstance(x, str):
        return (x,)
    return tuple(x)



import pandas as pd
from whitecanvas import new_canvas

def main():
    url = "https://raw.githubusercontent.com/mwaskom/seaborn-data/master/penguins.csv"
    df = pd.read_csv(url).dropna()

    canvas = new_canvas("matplotlib:qt")

    (
        canvas
        .cat(df, x="bill_length_mm", y="flipper_length_mm")
        .add_markers(color="species", size="bill_depth_mm")
    )

    canvas.add_legend()
    canvas.show(block=True)

if __name__ == "__main__":
    main()



from whitecanvas.backend.vispy.band import Band
from whitecanvas.backend.vispy.bars import Bars
from whitecanvas.backend.vispy.canvas import Canvas, CanvasGrid, as_overlay
from whitecanvas.backend.vispy.image import Image
from whitecanvas.backend.vispy.line import MonoLine, MultiLine
from whitecanvas.backend.vispy.markers import Markers
from whitecanvas.backend.vispy.text import Texts



from __future__ import annotations

from typing import Callable, cast

from qtpy import QtCore, QtGui
from qtpy import QtWidgets as QtW

from whitecanvas._axis import CategoricalAxis, DimAxis, RangeAxis
from whitecanvas.canvas import CanvasBase, CanvasGrid


class QtCanvas(QtW.QWidget):
    def __init__(self, canvas: CanvasGrid):
        super().__init__()
        self._canvas = canvas
        self._canvas_qimage = QtGui.QImage()

    def paintEvent(self, event: QtGui.QPaintEvent) -> None:
        rect = event.rect()
        painter = QtGui.QPainter(self)

        painter.eraseRect(rect)  # clear the widget canvas
        origin = QtCore.QPoint(rect.left(), rect.top())
        painter.drawImage(origin, self._canvas_qimage)

    def sizeHint(self) -> QtCore.QSize:
        w, h = self._canvas.size
        return QtCore.QSize(int(w), int(h))

    def _update_qimage(self):
        buf = self._canvas.screenshot()
        qimage = QtGui.QImage(
            buf.tobytes(),
            buf.shape[1],
            buf.shape[0],
            QtGui.QImage.Format.Format_RGBA8888,
        )

        qimage.setDevicePixelRatio(self.devicePixelRatioF())
        self._canvas_qimage = qimage

    def _update_widget_state(self):
        self._update_qimage()
        self.update()

    def resizeEvent(self, event: QtGui.QResizeEvent) -> None:
        qsize = event.size()

        self._canvas.size = qsize.width(), qsize.height()
        self._update_widget_state()


class QtMainWindow(QtW.QMainWindow):
    _instance = None

    def __init__(self, canvas: CanvasGrid):
        super().__init__()
        self.setWindowTitle("whitecanvas")
        self._widget = QtCanvas(canvas)
        self.setCentralWidget(self._widget)
        sl = QtDimSliders.from_canvas(canvas[0, 0], self)
        dock = QtW.QDockWidget("Dimensions", self)
        dock.setWidget(sl)
        self.addDockWidget(QtCore.Qt.DockWidgetArea.BottomDockWidgetArea, dock)
        canvas.events.drawn.connect(self._widget._update_widget_state, max_args=0)
        self.__class__._instance = self


class QtDimSliders(QtW.QWidget):
    changed = QtCore.Signal(object)

    def __init__(self, parent=None):
        super().__init__(parent)
        layout = QtW.QFormLayout()
        self.setLayout(layout)
        self._layout = layout
        self._widgets: dict[str, QtW.QWidget] = {}

    def set_axes(self, axes: list[DimAxis]):
        for _ in range(len(self._widgets)):
            self._layout.removeRow(0)
        self._widgets.clear()
        for ax in axes:
            if isinstance(ax, RangeAxis):
                widget = QtW.QSlider(QtCore.Qt.Orientation.Horizontal)
                widget.setRange(0, ax.size() - 1)
                widget.valueChanged.connect(self._emit_changed)
            elif isinstance(ax, CategoricalAxis):
                widget = QtW.QComboBox()
                widget.addItems(ax.categories())
                widget.currentIndexChanged.connect(self._emit_changed)
            else:
                raise NotImplementedError
            self._layout.addRow(ax.name, widget)
            self._widgets[ax.name] = widget

    def connect_changed(self, callback: Callable[[dict[str, object]], None]):
        self.changed.connect(callback)

    def _emit_changed(self):
        values = {}
        for name, widget in self._widgets.items():
            if isinstance(widget, QtW.QSlider):
                values[name] = cast(QtW.QSlider, widget).value()
            elif isinstance(widget, QtW.QComboBox):
                cbox = cast(QtW.QComboBox, widget)
                idx = cbox.currentIndex()
                values[name] = cbox.itemText(idx)
            else:
                raise NotImplementedError
        self.changed.emit(values)

    @classmethod
    def from_canvas(cls, canvas: CanvasBase, parent=None):
        self = cls(parent=parent)

        @canvas.dims.events.axis_names.connect
        def _update_axes():
            self.set_axes(canvas.dims._axes)

        self.connect_changed(canvas.dims.set_indices)
        _update_axes()
        return self



from whitecanvas.canvas._base import Canvas, CanvasBase
from whitecanvas.canvas._grid import (
    CanvasGrid,
    CanvasHGrid,
    CanvasVGrid,
    SingleCanvas,
)
from whitecanvas.canvas._joint import JointGrid
from whitecanvas.canvas._linker import link_axes

__all__ = [
    "CanvasBase",
    "Canvas",
    "CanvasGrid",
    "CanvasHGrid",
    "CanvasVGrid",
    "JointGrid",
    "SingleCanvas",
    "link_axes",
]



from __future__ import annotations

import weakref
from typing import TYPE_CHECKING

import numpy as np
from bokeh.models import FactorRange
from cmap import Color

from whitecanvas.backend.bokeh._base import to_bokeh_line_style
from whitecanvas.types import LineStyle

if TYPE_CHECKING:
    from bokeh.models import Axis as BokehAxis
    from bokeh.models import DataRange1d
    from bokeh.models import Grid as BokehGrid

    from whitecanvas.backend.bokeh.canvas import Canvas


class _CanvasComponent:
    def __init__(self, canvas: Canvas):
        self._canvas = weakref.ref(canvas)

    @property
    def _plot(self):
        return self._canvas()._plot


class Title(_CanvasComponent):
    def _plt_get_visible(self) -> bool:
        return self._plot.title.visible

    def _plt_set_visible(self, visible: bool):
        self._plot.title.visible = visible

    def _plt_get_text(self) -> str:
        return self._plot.title.text

    def _plt_set_text(self, text: str):
        self._plot.title.text = text

    def _plt_get_color(self):
        return np.array(Color(self._plot.title.text_color).rgba)

    def _plt_set_color(self, color):
        self._plot.title.text_color = Color(color).hex

    def _plt_get_size(self) -> int:
        return int(self._plot.title.text_font_size.rstrip("pt"))

    def _plt_set_size(self, size: int):
        self._plot.title.text_font_size = f"{size}pt"

    def _plt_get_fontfamily(self) -> str:
        return self._plot.title.text_font

    def _plt_set_fontfamily(self, family: str):
        self._plot.title.text_font = family


class X:
    def _plt_get_axis(self) -> BokehAxis:
        return self._canvas()._get_xaxis()

    def _plt_get_range(self) -> DataRange1d:
        return self._canvas()._get_xrange()


class Y:
    def _plt_get_axis(self) -> BokehAxis:
        return self._canvas()._get_yaxis()

    def _plt_get_range(self) -> DataRange1d:
        return self._canvas()._get_yrange()


class Axis(_CanvasComponent):
    def __init__(self, canvas: Canvas):
        super().__init__(canvas)

    def _plt_get_axis(self) -> BokehAxis:
        raise NotImplementedError

    def _plt_get_visible(self) -> bool:
        return self._plt_get_axis().visible

    def _plt_set_visible(self, visible: bool):
        self._plt_get_axis().visible = visible

    def _plt_get_color(self):
        return np.array(Color(self._plt_get_axis().axis_label_text_color).rgba)

    def _plt_set_color(self, color):
        self._plt_get_axis().axis_label_text_color = Color(color).hex


class Label(_CanvasComponent):
    def __init__(self, canvas: Canvas):
        super().__init__(canvas)
        self._text = ""
        self._visible = True

    def _plt_get_axis(self) -> BokehAxis:
        raise NotImplementedError

    def _plt_get_visible(self) -> bool:
        return self._visible

    def _plt_set_visible(self, visible: bool):
        self._visible = visible
        if not visible:
            self._plt_get_axis().axis_label = None
        else:
            self._plt_get_axis().axis_label = self._text

    def _plt_get_text(self) -> str:
        return self._text

    def _plt_set_text(self, text: str):
        self._text = text
        if self._plt_get_visible():
            self._plt_get_axis().axis_label = text

    def _plt_get_color(self):
        return np.array(Color(self._plt_get_axis().axis_label_text_color).rgba)

    def _plt_set_color(self, color):
        self._plt_get_axis().axis_label_text_color = Color(color).hex

    def _plt_get_size(self) -> int:
        return int(self._plt_get_axis().axis_label_text_font_size.rstrip("pt"))

    def _plt_set_size(self, size: int):
        self._plt_get_axis().axis_label_text_font_size = f"{size}pt"

    def _plt_get_fontfamily(self) -> str:
        return self._plt_get_axis().axis_label_text_font

    def _plt_set_fontfamily(self, family: str):
        self._plt_get_axis().axis_label_text_font = family


class Ticks(_CanvasComponent):
    def __init__(self, canvas: Canvas):
        super().__init__(canvas)
        self._visible = True
        self._font_size = None
        self._tick_color = None

    def _plt_get_axis(self) -> BokehAxis:
        raise NotImplementedError

    def _plt_get_tick_labels(self) -> tuple[list[float], list[str]]:
        return [], []  # Not implemented on bokeh side

    def _plt_override_labels(self, pos: list[float], labels: list[str]):
        self._plt_get_axis().ticker = pos
        self._plt_get_axis().major_label_overrides = dict(zip(pos, labels))

    def _plt_reset_override(self):
        self._plt_get_axis().ticker = []
        self._plt_get_axis().major_label_overrides = {}

    def _plt_get_visible(self) -> bool:
        return self._visible

    def _plt_set_visible(self, visible: bool):
        self._visible = visible
        if not visible:
            self._plt_get_axis().major_label_text_font_size = "0pt"
            self._plt_get_axis().major_tick_line_color = None
            self._plt_get_axis().minor_tick_line_color = None
        else:
            self._plt_get_axis().major_label_text_font_size = self._font_size
            self._plt_get_axis().major_tick_line_color = self._tick_color
            self._plt_get_axis().minor_tick_line_color = self._tick_color

    def _plt_get_size(self) -> float:
        return self._font_size

    def _plt_set_size(self, size: int):
        self._font_size = f"{size}pt"
        if self._visible:
            self._plt_get_axis().major_label_text_font_size = self._font_size

    def _plt_get_fontfamily(self) -> str:
        return self._plt_get_axis().major_label_text_font

    def _plt_set_fontfamily(self, font):
        self._plt_get_axis().major_label_text_font = font

    def _plt_get_color(self):
        return np.array(Color(self._tick_color).rgba)

    def _plt_set_color(self, color):
        self._tick_color = Color(color).hex
        if self._visible:
            self._plt_get_axis().major_label_text_color = self._tick_color
            self._plt_get_axis().major_tick_line_color = self._tick_color
            self._plt_get_axis().minor_tick_line_color = self._tick_color

    def _plt_get_text_rotation(self) -> float:
        return self._plt_get_axis().major_label_orientation

    def _plt_set_text_rotation(self, rotation: float):
        self._plt_get_axis().major_label_orientation = rotation


class XAxis(X, Axis):
    def _plt_set_grid_state(self, visible: bool, color, width: float, style: LineStyle):
        grid: BokehGrid = self._canvas()._plot.xgrid
        grid.visible = visible
        grid.grid_line_color = Color(color).hex
        grid.grid_line_width = width
        grid.grid_line_dash = to_bokeh_line_style(style)

    def _plt_flip(self) -> None:
        fig = self._canvas()._plot
        fig.x_range.flipped = not fig.x_range.flipped

    def _plt_get_limits(self) -> tuple[float, float]:
        _range = self._plt_get_range()
        return _range.start, _range.end

    def _plt_set_limits(self, limits: tuple[float, float]):
        _range = self._plt_get_range()
        _range.start, _range.end = limits


class YAxis(Y, Axis):
    def _plt_set_grid_state(self, visible: bool, color, width: float, style: LineStyle):
        grid: BokehGrid = self._canvas()._plot.ygrid
        grid.visible = visible
        grid.grid_line_color = Color(color).hex
        grid.grid_line_width = width
        grid.grid_line_dash = to_bokeh_line_style(style)

    def _plt_flip(self) -> None:
        fig = self._canvas()._plot
        fig.y_range.flipped = not fig.y_range.flipped

    def _plt_get_limits(self) -> tuple[float, float]:
        limits = self._plt_get_range()
        return limits.start, limits.end

    def _plt_set_limits(self, limits: tuple[float, float]):
        x_range = self._plt_get_range()
        x_range.start, x_range.end = limits


class XLabel(X, Label):
    pass


class YLabel(Y, Label):
    pass


class XTicks(X, Ticks):
    def _plt_set_multilevel_text(self, text: list[tuple[str, ...]]):
        self._canvas()._plot.x_range = FactorRange(*text)


class YTicks(Y, Ticks):
    def _plt_set_multilevel_text(self, text: list[tuple[str, ...]]):
        self._canvas()._plot.y_range = FactorRange(*text)



from __future__ import annotations

import tkinter as tk
from tkinter import ttk
from typing import Callable

import numpy as np
from PIL import Image, ImageTk
from psygnal import Signal

from whitecanvas._axis import CategoricalAxis, DimAxis, RangeAxis
from whitecanvas.canvas import CanvasBase, CanvasGrid


class TkCanvas(ttk.Frame):
    def __init__(self, canvas: CanvasGrid):
        super().__init__()
        tkc = tk.Canvas(width=300, height=300)
        tkc.pack()
        self._tkcanvas = tkc
        self._canvas = canvas
        self._canvas_qimage = ImageTk.PhotoImage(
            image=Image.fromarray(np.zeros((5, 5, 3), dtype=np.uint8))
        )

    def resize(self, event: tk.Event):
        if event.widget.widgetName == "toplevel":
            w, h = event.width, event.height
            self._canvas.size = w, h
            self._update_imagetk()
            self.update()

    def _update_imagetk(self):
        buf = self._canvas.screenshot()
        self._canvas_qimage = ImageTk.PhotoImage(image=Image.fromarray(buf))
        self._tkcanvas.create_image(20, 20, anchor="nw", image=self._canvas_qimage)


class TkMainWindow(ttk.Frame):
    _instance = None

    def __init__(self, canvas: CanvasGrid):
        super().__init__()
        tkcanvas = TkCanvas(canvas)
        tkcanvas.pack(fill="both", expand=True)
        sl = TkDimSliders(self)
        sl.pack(fill="both", expand=True)
        self.__class__._instance = self
        canvas.events.drawn.connect(tkcanvas._update_imagetk, max_args=0)


class TkSlider(ttk.Scale):
    changed = Signal(int)

    def __init__(self, parent=None, **kwargs):
        kwargs["command"] = lambda value: self.changed.emit(int(value))
        super().__init__(parent, **kwargs)


class TkComboBox(ttk.Combobox):
    changed = Signal(str)

    def __init__(self, parent=None, **kwargs):
        kwargs["postcommand"] = lambda: self.changed.emit(self.get())
        super().__init__(parent, **kwargs)


class TkDimSliders(ttk.Frame):
    changed = Signal(dict)

    def __init__(self, parent=None):
        super().__init__(parent)
        self._layout = ttk.Frame(self)
        self._layout.pack(fill="both", expand=True)
        self._widgets = {}

    def set_axes(self, axes: list[DimAxis]):
        for axis in axes:
            if isinstance(axis, RangeAxis):
                slider = TkSlider(
                    self._layout,
                    from_=0,
                    to=axis.size(),
                    value=axis.value(),
                    orient="horizontal",
                )
                slider.pack(fill="both", expand=True)
            elif isinstance(axis, CategoricalAxis):
                slider = TkComboBox(
                    self._layout,
                    values=axis.categories(),
                    # state="readonly",
                    textvariable=axis.value(),
                )
                slider.pack(fill="both", expand=True)
            else:
                raise ValueError(f"Unknown axis type {axis}")
            self._widgets[axis.name] = slider

    def connect_changed(self, callback: Callable[[dict[str, object]], None]):
        self.changed.connect(callback, max_args=1)

    def _emit_changed(self):
        values = {}
        for name, widget in self._widgets.items():
            if isinstance(widget, TkSlider):
                values[name] = int(widget.get())
            elif isinstance(widget, TkComboBox):
                values[name] = widget.get()
            else:
                raise NotImplementedError
        self.changed.emit(values)

    def show(self):
        self.mainloop()

    @classmethod
    def from_canvas(cls, canvas: CanvasBase, parent=None):
        self = cls(parent=parent)

        @canvas.dims.events.axis_names.connect
        def _update_axes():
            self.set_axes(canvas.dims._axes)

        self.connect_changed(canvas.dims.set_indices)
        _update_axes()
        return self



from __future__ import annotations

import numpy as np
from vispy.scene import visuals

from whitecanvas.backend.vispy.markers import Markers as Markers2D


class Markers3D(Markers2D):
    def __init__(self, xdata, ydata, zdata):
        pos = np.stack([xdata, ydata, zdata], axis=1)
        visuals.Markers.__init__(self, pos=pos, edge_width=0, face_color="blue")
        self.unfreeze()
        self._hover_texts: list[str] | None = None

    ##### XYDataProtocol #####
    def _plt_get_data(self):
        data = self._data["a_position"]
        return data[:, 0], data[:, 1], data[:, 2]

    def _plt_set_data(self, xdata, ydata, zdata):
        ndata = self._plt_get_ndata()
        size = self._plt_get_symbol_size()
        edge_width = self._plt_get_edge_width()
        face_color = self._plt_get_face_color()
        edge_color = self._plt_get_edge_color()
        if xdata.size > ndata:
            size = np.concatenate([size, np.full(xdata.size - ndata, size[-1])])
            edge_width = np.concatenate(
                [edge_width, np.full(xdata.size - ndata, edge_width[-1])]
            )
            face_color = np.concatenate(
                [face_color, np.full((xdata.size - ndata, 4), face_color[-1])]
            )
            edge_color = np.concatenate(
                [edge_color, np.full((xdata.size - ndata, 4), edge_color[-1])]
            )
        elif xdata.size < ndata:
            size = size[: xdata.size]
            edge_width = edge_width[: xdata.size]
            face_color = face_color[: xdata.size]
            edge_color = edge_color[: xdata.size]
        self.set_data(
            pos=np.stack([xdata, ydata, zdata], axis=1),
            size=size,
            edge_width=edge_width,
            face_color=face_color,
            edge_color=edge_color,
            symbol=self.symbol[0],
        )

    def _plt_set_hover_text(self, text: list[str]):
        # TODO: not used yet
        self._hover_texts = text

    def _compute_bounds(self, axis, view):
        # override to fix the bounds computation
        pos = self._data["a_position"]
        if pos is None or pos.size == 0:
            return None
        if pos.shape[1] > axis:
            return (pos[:, axis].min(), pos[:, axis].max())
        else:
            return (0, 0)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Generic, Sequence, TypeVar

import numpy as np
from numpy.typing import ArrayLike, NDArray

from whitecanvas.backend import Backend
from whitecanvas.layers import _legend
from whitecanvas.layers._base import DataBoundLayer
from whitecanvas.layers._mixin import (
    ConstEdge,
    ConstFace,
    EdgeNamespace,
    FaceNamespace,
    MultiEdge,
    MultiFace,
    MultiFaceEdgeMixin,
)
from whitecanvas.protocols import BarProtocol
from whitecanvas.types import (
    ColorType,
    Hatch,
    LineStyle,
    Orientation,
    OrientationLike,
    Rect,
    _Void,
)

if TYPE_CHECKING:
    from whitecanvas.canvas import Canvas

_Face = TypeVar("_Face", bound=FaceNamespace)
_Edge = TypeVar("_Edge", bound=EdgeNamespace)
_void = _Void()


class Spans(
    DataBoundLayer[BarProtocol, NDArray[np.number]],
    MultiFaceEdgeMixin[_Face, _Edge],
    Generic[_Face, _Edge],
):
    """
    Layer that represents vertical/hosizontal spans.

    Attributes
    ----------
    face : `whitecanvas.layers._mixin.FaceNamespace`
        Face properties of the spans.
    edge : `whitecanvas.layers._mixin.EdgeNamespace`
        Edge properties of the spans.
    """

    _backend_class_name = "Bars"
    _NO_PADDING_NEEDED = True

    def __init__(
        self,
        spans: ArrayLike,
        *,
        name: str | None = None,
        orient: OrientationLike = "vertical",
        color: ColorType = "blue",
        alpha: float = 1.0,
        hatch: str | Hatch = Hatch.SOLID,
        backend: Backend | str | None = None,
    ):
        MultiFaceEdgeMixin.__init__(self)
        _spans = self._norm_layer_data(spans)
        ori = Orientation.parse(orient)
        nspans = _spans.shape[0]
        if ori.is_vertical:
            xxyy = _spans[:, 0], _spans[:, 1], np.zeros(nspans), np.ones(nspans)
            xhint = _spans.min(), _spans.max()
            yhint = None
        else:
            xxyy = np.zeros(nspans), np.ones(nspans), _spans[:, 0], _spans[:, 1]
            xhint = None
            yhint = _spans.min(), _spans.max()
        super().__init__(name=name)
        self._backend = self._create_backend(Backend(backend), *xxyy)
        self._orient = ori
        self.face.update(color=color, alpha=alpha, hatch=hatch)
        self._x_hint, self._y_hint = xhint, yhint
        self._low_lim = -1e10
        self._high_lim = 1e10
        self._init_events()

    @property
    def orient(self) -> Orientation:
        """Orientation of the spans."""
        return self._orient

    def _get_layer_data(self) -> NDArray[np.number]:
        x0, x1, y0, y1 = self._backend._plt_get_data()
        if self.orient.is_vertical:
            return np.column_stack([x0, x1])
        else:
            return np.column_stack([y0, y1])

    def _norm_layer_data(self, data: Any) -> NDArray[np.number]:
        _spans = np.asarray(data)
        if _spans.ndim != 2:
            raise ValueError(f"spans must be 2-dimensional, got {_spans.ndim}")
        if _spans.shape[1] != 2:
            raise ValueError(f"spans must be (N, 2), got {_spans.shape}")
        if _spans.dtype.kind not in "uif":
            raise ValueError(f"spans must be numeric, got {_spans.dtype}")
        return _spans

    def _set_layer_data(self, data: NDArray[np.number]):
        _low = np.full((data.shape[0],), self._low_lim)
        _high = np.full((data.shape[0],), self._high_lim)
        if self.orient.is_vertical:
            xxyy = data[:, 0], data[:, 1], _low, _high
            self._x_hint = data.min(), data.max()
        else:
            xxyy = _low, _high, data[:, 0], data[:, 1]
            self._y_hint = data.min(), data.max()
        self._backend._plt_set_data(*xxyy)

    def set_data(self, spans: ArrayLike):
        self.data = spans

    @property
    def ndata(self) -> int:
        """The number of data points"""
        return self._backend._plt_get_data()[0].size

    def _connect_canvas(self, canvas: Canvas):
        canvas.events.lims.connect(self._recalculate_spans, max_args=1)
        self._force_update_spans()
        return super()._connect_canvas(canvas)

    def _disconnect_canvas(self, canvas: Canvas):
        canvas.events.lims.connect(self._recalculate_spans, max_args=1)
        return super()._disconnect_canvas(canvas)

    def _recalculate_spans(self, rect: Rect):
        # update the rectangles so that their limits are not visible
        if self.orient.is_vertical:
            _min, _max = rect.bottom, rect.top
        else:
            _min, _max = rect.left, rect.right
        if _min > _max:
            _min, _max = _max, _min
        if _min >= self._low_lim and _max <= self._high_lim:
            return
        self._low_lim = _min - 1e10
        self._high_lim = _max + 1e10
        self._force_update_spans()

    def _force_update_spans(self):
        x0, x1, y0, y1 = self._backend._plt_get_data()
        _min_arr = np.full(x0.shape, self._low_lim)
        _max_arr = np.full(x0.shape, self._high_lim)
        if self.orient.is_vertical:
            spans = x0, x1, _min_arr, _max_arr
        else:
            spans = _min_arr, _max_arr, y0, y1
        self._backend._plt_set_data(*spans)

    def with_face(
        self,
        color: ColorType | _Void = _void,
        hatch: Hatch | str = Hatch.SOLID,
        alpha: float = 1,
    ) -> Spans[ConstFace, _Edge]:
        """Update the face properties."""
        return super().with_face(color, hatch, alpha)

    def with_face_multi(
        self,
        color: ColorType | Sequence[ColorType] | _Void = _void,
        hatch: str | Hatch | Sequence[str | Hatch] = Hatch.SOLID,
        alpha: float = 1,
    ) -> Spans[MultiFace, _Edge]:
        return super().with_face_multi(color, hatch, alpha)

    def with_edge(
        self,
        color: ColorType | None = None,
        width: float = 1,
        style: LineStyle | str = LineStyle.SOLID,
        alpha: float = 1,
    ) -> Spans[_Face, ConstEdge]:
        """Update the edge properties."""
        return super().with_edge(color, width, style, alpha)

    def with_edge_multi(
        self,
        color: ColorType | Sequence[ColorType] | None = None,
        width: float | Sequence[float] = 1,
        style: str | LineStyle | list[str | LineStyle] = LineStyle.SOLID,
        alpha: float = 1,
    ) -> Spans[_Face, MultiEdge]:
        return super().with_edge_multi(color, width, style, alpha)

    def _as_legend_item(self) -> _legend.BarLegendItem:
        return _legend.BarLegendItem(
            self.face._as_legend_info(), self.edge._as_legend_info()
        )



from functools import singledispatch
from typing import Generic, TypeVar

import pyqtgraph as pg
from pyqtgraph.graphicsItems.ScatterPlotItem import drawSymbol
from qtpy import QtCore, QtGui

from whitecanvas.backend.pyqtgraph._qt_utils import (
    array_to_qcolor,
    to_qt_brush_style,
    to_qt_line_style,
    to_qt_symbol,
)
from whitecanvas.layers import _legend as _leg

_L = TypeVar("_L", bound=_leg.LegendItem)


@singledispatch
def make_sample_item(item) -> "QtItemSampleBase | None":
    return None


@make_sample_item.register
def _(item: _leg.LineLegendItem):
    return LineItemSample(item)


@make_sample_item.register
def _(item: _leg.MarkersLegendItem):
    return MarkerItemSample(item)


@make_sample_item.register
def _(item: _leg.BarLegendItem):
    return BarItemSample(item)


@make_sample_item.register
def _(item: _leg.PlotLegendItem):
    return PlotItemSample(item)


@make_sample_item.register
def _(item: _leg.ErrorLegendItem):
    return ErrorItemSample(item)


@make_sample_item.register
def _(item: _leg.LineErrorLegendItem):
    return LineErrorItemSample(item)


@make_sample_item.register
def _(item: _leg.MarkerErrorLegendItem):
    return MarkerErrorItemSample(item)


@make_sample_item.register
def _(item: _leg.PlotErrorLegendItem):
    return PlotErrorItemSample(item)


@make_sample_item.register
def _(item: _leg.StemLegendItem):
    return StemItemSample(item)


@make_sample_item.register
def _(item: _leg.TitleItem):
    return TitleItemSample(item)


class QtItemSampleBase(pg.GraphicsWidget, Generic[_L]):
    """Class responsible for drawing a single item in a LegendItem (sans label)"""

    def __init__(self, item: _L):
        pg.GraphicsWidget.__init__(self)
        self._item = item

    def boundingRect(self):
        return QtCore.QRectF(0, 0, 20, 20)

    def paint(self, p, *args):
        pass


class LineItemSample(QtItemSampleBase[_leg.LineLegendItem]):
    def paint(self, p: QtGui.QPainter, *args):
        p.setPen(_edge_to_qpen(self._item))
        p.drawLine(0, 11, 20, 11)


class MarkerItemSample(QtItemSampleBase[_leg.MarkersLegendItem]):
    def paint(self, p: QtGui.QPainter, *args):
        p.translate(10, 10)
        drawSymbol(
            p,
            to_qt_symbol(self._item.symbol),
            self._item.size,
            _edge_to_qpen(self._item.edge),
            _face_to_qbrush(self._item.face),
        )


class BarItemSample(QtItemSampleBase[_leg.BarLegendItem]):
    def paint(self, p: QtGui.QPainter, *args):
        p.setBrush(_face_to_qbrush(self._item))
        p.setPen(_edge_to_qpen(self._item))
        p.drawRect(0, 4, 20, 12)


class PlotItemSample(QtItemSampleBase[_leg.PlotLegendItem]):
    def paint(self, p: QtGui.QPainter, *args):
        p.setPen(_edge_to_qpen(self._item.line))
        p.drawLine(0, 11, 20, 11)
        p.translate(10, 10)
        drawSymbol(
            p,
            to_qt_symbol(self._item.markers.symbol),
            self._item.markers.size,
            _edge_to_qpen(self._item.markers.edge),
            _face_to_qbrush(self._item.markers.face),
        )


class ErrorItemSample(QtItemSampleBase[_leg.ErrorLegendItem]):
    def paint(self, p: QtGui.QPainter, *args):
        p.setPen(_edge_to_qpen(self._item))
        p.drawLine(10, 4, 10, 16)
        if self._item.capsize > 0:
            p.drawLine(8, 4, 12, 4)
            p.drawLine(8, 16, 12, 16)


class LineErrorItemSample(QtItemSampleBase[_leg.LineErrorLegendItem]):
    def paint(self, p: QtGui.QPainter, *args):
        p.setPen(_edge_to_qpen(self._item.line))
        p.drawLine(0, 11, 20, 11)
        if self._item.xerr is not None:
            err = self._item.xerr
        elif self._item.yerr is not None:
            err = self._item.yerr
        else:
            return
        p.setPen(_edge_to_qpen(err))
        p.drawLine(10, 4, 10, 16)
        if err.capsize > 0:
            p.drawLine(8, 4, 12, 4)
            p.drawLine(8, 16, 12, 16)


class MarkerErrorItemSample(QtItemSampleBase[_leg.MarkerErrorLegendItem]):
    def paint(self, p: QtGui.QPainter, *args):
        p.translate(10, 10)
        drawSymbol(
            p,
            to_qt_symbol(self._item.markers.symbol),
            self._item.markers.size,
            _edge_to_qpen(self._item.markers.edge),
            _face_to_qbrush(self._item.markers.face),
        )
        if self._item.xerr is not None:
            p.setPen(_edge_to_qpen(self._item.xerr))
            p.drawLine(4, 10, 16, 10)
            if self._item.xerr.capsize > 0:
                p.drawLine(4, 8, 4, 12)
                p.drawLine(16, 8, 16, 12)
        if self._item.yerr is not None:
            p.setPen(_edge_to_qpen(self._item.yerr))
            p.drawLine(10, 4, 10, 16)
            if self._item.yerr.capsize > 0:
                p.drawLine(8, 4, 12, 4)
                p.drawLine(8, 16, 12, 16)


class PlotErrorItemSample(QtItemSampleBase[_leg.PlotErrorLegendItem]):
    def paint(self, p: QtGui.QPainter, *args):
        p.setPen(_edge_to_qpen(self._item.plot.line))
        p.drawLine(0, 11, 20, 11)
        p.translate(10, 10)
        drawSymbol(
            p,
            to_qt_symbol(self._item.plot.markers.symbol),
            self._item.plot.markers.size,
            _edge_to_qpen(self._item.plot.markers.edge),
            _face_to_qbrush(self._item.plot.markers.face),
        )
        if self._item.xerr is not None:
            p.setPen(_edge_to_qpen(self._item.xerr))
            p.drawLine(4, 10, 16, 10)
            if self._item.xerr.capsize > 0:
                p.drawLine(4, 8, 4, 12)
                p.drawLine(16, 8, 16, 12)
        if self._item.yerr is not None:
            p.setPen(_edge_to_qpen(self._item.yerr))
            p.drawLine(10, 4, 10, 16)
            if self._item.yerr.capsize > 0:
                p.drawLine(8, 4, 12, 4)
                p.drawLine(8, 16, 12, 16)


class StemItemSample(QtItemSampleBase[_leg.StemLegendItem]):
    def paint(self, p: QtGui.QPainter, *args):
        p.setPen(_edge_to_qpen(self._item.line))
        p.drawLine(10, 0, 10, 12)
        p.translate(10, 12)
        drawSymbol(
            p,
            to_qt_symbol(self._item.markers.symbol),
            self._item.markers.size,
            _edge_to_qpen(self._item.markers.edge),
            _face_to_qbrush(self._item.markers.face),
        )


class TitleItemSample(QtItemSampleBase[_leg.TitleItem]):
    def paint(self, p: QtGui.QPainter, *args):
        pass


def _face_to_qbrush(face: _leg.FaceInfo) -> QtGui.QBrush:
    brush = QtGui.QBrush()
    brush.setColor(array_to_qcolor(face.color))
    brush.setStyle(to_qt_brush_style(face.hatch))
    return brush


def _edge_to_qpen(edge: _leg.EdgeInfo) -> QtGui.QPen:
    pen = QtGui.QPen()
    pen.setCosmetic(True)
    pen.setColor(array_to_qcolor(edge.color))
    pen.setStyle(to_qt_line_style(edge.style))
    pen.setWidthF(edge.width)
    return pen



from __future__ import annotations

from typing import NamedTuple

import numpy as np
from numpy.typing import NDArray


class XYData(NamedTuple):
    """
    Tuple of x and y array.

    Used for data of Line, Markers etc.
    """

    x: NDArray[np.floating]
    y: NDArray[np.floating]

    def stack(self) -> NDArray[np.floating]:
        """Data as a stacked (N, 2) array."""
        return np.stack([self.x, self.y], axis=1)


class XYYData(NamedTuple):
    """
    Tuple of x, y0, and y1 array.

    Used for data of Bars, Errorbars etc.
    """

    x: NDArray[np.floating]
    y0: NDArray[np.floating]
    y1: NDArray[np.floating]

    @property
    def ycenter(self) -> NDArray[np.floating]:
        return (self.y0 + self.y1) / 2

    @property
    def ydiff(self) -> NDArray[np.floating]:
        return self.y1 - self.y0


class XYTextData(NamedTuple):
    x: NDArray[np.floating]
    y: NDArray[np.floating]
    text: NDArray[np.object_]


class XYZData(NamedTuple):
    """
    Tuple of x, y, and z array.

    Used for data of Surface etc.
    """

    x: NDArray[np.floating]
    y: NDArray[np.floating]
    z: NDArray[np.floating]

    def stack(self) -> NDArray[np.floating]:
        """Data as a stacked (N, 3) array."""
        return np.stack([self.x, self.y, self.z], axis=1)


class XYVectorData(NamedTuple):
    x: NDArray[np.floating]
    y: NDArray[np.floating]
    vx: NDArray[np.floating]
    vy: NDArray[np.floating]


class XYZVectorData(NamedTuple):
    x: NDArray[np.floating]
    y: NDArray[np.floating]
    z: NDArray[np.floating]
    vx: NDArray[np.floating]
    vy: NDArray[np.floating]
    vz: NDArray[np.floating]


class MeshData(NamedTuple):
    vertices: NDArray[np.floating]
    faces: NDArray[np.intp]


class Rect(NamedTuple):
    """Rectangular range in left, right, bottom, top order."""

    left: float
    right: float
    bottom: float
    top: float

    def __repr__(self) -> str:
        return (
            f"Rect(left={self.left:.6g}, right={self.right:.6g}, "
            f"bottom={self.bottom:.6g}, top={self.top:.6g})"
        )

    @classmethod
    def with_check(cls, left: float, right: float, bottom: float, top: float):
        if left > right:
            raise ValueError("left must be less than or equal to right")
        if bottom > top:
            raise ValueError("bottom must be less than or equal to top")
        return cls(float(left), float(right), float(bottom), float(top))

    @classmethod
    def with_sort(cls, left: float, right: float, bottom: float, top: float):
        left, right = sorted([left, right])
        bottom, top = sorted([bottom, top])
        return cls(float(left), float(right), float(bottom), float(top))

    @property
    def width(self) -> float:
        """Width of the range."""
        return self.right - self.left

    @property
    def height(self) -> float:
        """Height of the range."""
        return self.top - self.bottom

    @property
    def size(self) -> tuple[float, float]:
        """Size (width, height) of the range."""
        return self.width, self.height

    @property
    def center(self) -> tuple[float, float]:
        """Center of the range."""
        return (self.left + self.right) / 2, (self.top + self.bottom) / 2



from __future__ import annotations

import numpy as np
from numpy.typing import NDArray
from vispy.color import Color as vispyColor
from vispy.gloo import set_state
from vispy.scene import visuals

from whitecanvas.backend import _not_implemented
from whitecanvas.protocols import BandProtocol, check_protocol
from whitecanvas.types import LineStyle, Orientation


# vispy's Polygon is not well implemented. Use custom class.
class VispyBand(visuals.Compound):
    def __init__(
        self,
        pos: NDArray[np.floating],
        color="black",
        border_color=None,
        border_width=1,
        **kwargs,
    ):
        self._mesh = visuals.Mesh()
        self._border = visuals.Line(method="gl")
        self._pos = pos
        self._color = vispyColor(color)
        self._border_width = border_width
        self._border_color = vispyColor(border_color)

        self._update()
        visuals.Compound.__init__(self, [self._mesh, self._border], **kwargs)
        self._mesh.set_gl_state(
            polygon_offset_fill=True,
            polygon_offset=(1, 1),
        )
        self.freeze()

    def _update(self):
        set_state(polygon_offset_fill=False)
        npos = self._pos.shape[0]
        _index = np.arange(npos // 2 - 1, dtype=np.uint32)
        _index_inv = npos - _index
        faces0 = np.stack([_index, _index + 1, _index_inv - 1], axis=1)
        faces1 = np.stack([_index + 1, _index_inv - 1, _index_inv - 2], axis=1)
        faces = np.concatenate([faces0, faces1], axis=0)
        self._mesh.set_data(vertices=self._pos, faces=faces, color=self._color.rgba)
        if not self._border_color.is_blank:
            # Close border if it is not already.
            border_pos = self._pos
            if np.any(border_pos[0] != border_pos[-1]):
                border_pos = np.concatenate([border_pos, border_pos[:1]], axis=0)
            self._border.set_data(
                pos=border_pos, color=self._border_color.rgba, width=self._border_width
            )

            self._border.update()

    def set_pos(self, pos: NDArray[np.floating]):
        self._pos = pos
        self._update()

    @property
    def color(self):
        """The color of the polygon."""
        return self._color

    @color.setter
    def color(self, color):
        self._color = vispyColor(color, clip=True)
        self._mesh.set_data(
            vertices=self._mesh.mesh_data.get_vertices(),
            faces=self._mesh.mesh_data.get_faces(),
            color=self._color.rgba,
        )

    @property
    def border_color(self):
        """The border color of the polygon."""
        return self._border_color

    @border_color.setter
    def border_color(self, border_color):
        self._border_color = vispyColor(border_color)
        self._update()

    @property
    def border_width(self) -> float:
        return self._border.width

    @border_width.setter
    def border_width(self, width: float):
        self._border_width = width
        self._update()


@check_protocol(BandProtocol)
class Band(VispyBand):
    def __init__(self, t, ydata0, ydata1, orient: Orientation):
        if orient.is_vertical:
            fw = np.stack([t, ydata0], axis=1)
            bw = np.stack([t[::-1], ydata1[::-1]], axis=1)
        else:
            fw = np.stack([ydata0, t], axis=1)
            bw = np.stack([ydata1[::-1], t[::-1]], axis=1)
        verts = np.concatenate([fw, bw], axis=0)
        self._edge_style = LineStyle.SOLID
        super().__init__(verts, border_width=0)
        self.unfreeze()
        self._t = t
        self._y0 = ydata0
        self._y1 = ydata1

    ##### LayerProtocol #####
    def _plt_get_visible(self) -> bool:
        return self.visible

    def _plt_set_visible(self, visible: bool):
        self.visible = visible

    ##### XYDataProtocol #####
    def _plt_get_vertical_data(self):
        return self._t, self._y0, self._y1

    def _plt_get_horizontal_data(self):
        return self._t, self._y0, self._y1

    def _plt_set_vertical_data(self, t, ydata0, ydata1):
        verts = np.concatenate(
            [
                np.stack([t, ydata0], axis=1),
                np.stack([t[::-1], ydata1[::-1]], axis=1),
            ],
            axis=0,
        )
        self.set_pos(verts)
        self._t = t
        self._y0 = ydata0
        self._y1 = ydata1

    def _plt_set_horizontal_data(self, t, ydata0, ydata1):
        verts = np.concatenate(
            [
                np.stack([ydata0, t], axis=1),
                np.stack([ydata1[::-1], t[::-1]], axis=1),
            ],
            axis=0,
        )
        self.set_pos(verts)
        self._t = t
        self._y0 = ydata0
        self._y1 = ydata1

    ##### HasFace protocol #####
    def _plt_get_face_color(self) -> NDArray[np.float32]:
        return np.array(self.color, dtype=np.float32)

    def _plt_set_face_color(self, color: NDArray[np.float32]):
        self.color = color

    _plt_get_face_hatch, _plt_set_face_hatch = _not_implemented.face_hatch()

    ##### HasEdges protocol #####
    def _plt_get_edge_color(self) -> NDArray[np.float32]:
        return np.array(self.border_color, dtype=np.float32)

    def _plt_set_edge_color(self, color: NDArray[np.float32]):
        self.border_color = color

    def _plt_get_edge_width(self) -> float:
        return self.border_width

    def _plt_set_edge_width(self, width: float):
        self.border_width = width

    _plt_get_edge_style, _plt_set_edge_style = _not_implemented.edge_style()

    def _plt_set_hover_text(self, text: list[str]):
        # TODO: not used yet
        self._hover_texts = text

    def _plt_connect_pick_event(self, callback):
        pass



from __future__ import annotations

import numpy as np
from matplotlib.artist import Artist
from matplotlib.text import Text as mplText
from numpy.typing import NDArray

from whitecanvas.backend.matplotlib._base import MplLayer
from whitecanvas.protocols import TextProtocol, check_protocol
from whitecanvas.types import Alignment, Hatch, LineStyle
from whitecanvas.utils.normalize import as_color_array
from whitecanvas.utils.type_check import is_real_number


@check_protocol(TextProtocol)
class Texts(Artist, MplLayer):
    def __init__(
        self, x: NDArray[np.floating], y: NDArray[np.floating], text: list[str]
    ):
        super().__init__()
        self._children: list[mplText] = []
        for x0, y0, text0 in zip(x, y, text):
            self._children.append(
                mplText(
                    x0, y0, text0, verticalalignment="baseline",
                    horizontalalignment="left", clip_on=True,
                    color=np.array([0, 0, 0, 1], dtype=np.float32),
                )  # fmt: skip
            )
        self._font_family = "Arial"
        self._align = Alignment.BOTTOM_LEFT
        self._remove_method = _remove_method

    def draw(self, renderer):
        for child in self.get_children():
            child.draw(renderer)

    def get_children(self) -> list[mplText]:
        return self._children

    def set_visible(self, visible: bool):
        for child in self.get_children():
            child.set_visible(visible)
        super().set_visible(visible)

    def set_transform(self, transform):
        for child in self.get_children():
            child.set_transform(transform)
        super().set_transform(transform)

    def _plt_get_text(self) -> list[str]:
        return [child.get_text() for child in self.get_children()]

    def _plt_set_text(self, text: list[str]):
        for child, text0 in zip(self.get_children(), text):
            child.set_text(str(text0))

    def _plt_get_text_color(self):
        return np.stack([child.get_color() for child in self.get_children()], axis=0)

    def _plt_set_text_color(self, color):
        color = as_color_array(color, len(self.get_children()))
        for child, color0 in zip(self.get_children(), color):
            child.set_color(color0)

    def _plt_get_text_size(self) -> NDArray[np.float32]:
        return np.array(
            [child.get_fontsize() for child in self.get_children()], dtype=np.float32
        )

    def _plt_set_text_size(self, size: float | NDArray[np.float32]):
        if is_real_number(size):
            _size = np.full(len(self.get_children()), size, dtype=np.float32)
        else:
            _size = size
        for child, size0 in zip(self.get_children(), _size):
            child.set_fontsize(size0)

    def _plt_get_text_position(self) -> tuple[NDArray[np.float32], NDArray[np.float32]]:
        ar = np.array([child.get_position() for child in self.get_children()])
        if ar.size == 0:
            return np.array([]), np.array([])
        return ar[:, 0], ar[:, 1]

    def _plt_set_text_position(
        self, position: tuple[NDArray[np.float32], NDArray[np.float32]]
    ):
        x, y = position
        if x.size > len(self._children):
            for _ in range(x.size - len(self._children)):
                self._children.append(
                    mplText(
                        0, 0, "", verticalalignment="baseline",
                        horizontalalignment="left", clip_on=True,
                        color=np.array([0, 0, 0, 1], dtype=np.float32),
                    )  # fmt: skip
                )
        elif x.size < len(self._children):
            for c in self._children[x.size :]:
                c.remove()
            self._children = self._children[: x.size]
        for child, x0, y0 in zip(self.get_children(), x, y):
            child.set_position((x0, y0))

    def _plt_get_text_anchor(self) -> Alignment:
        return self._align

    def _plt_set_text_anchor(self, anc: Alignment):
        """Set the text position."""
        v, h = anc.split()
        va = _VERTICAL_ALIGNMENTS_INV[v]
        ha = _HORIZONTAL_ALIGNMENTS_INV[h]
        for child in self.get_children():
            child.set_verticalalignment(va)
            child.set_horizontalalignment(ha)
        self._align = anc

    def _plt_get_text_rotation(self) -> NDArray[np.float32]:
        return np.array(
            [child.get_rotation() for child in self.get_children()], dtype=np.float32
        )

    def _plt_set_text_rotation(self, rotation: NDArray[np.float32]):
        for child, rotation0 in zip(self.get_children(), rotation):
            child.set_rotation(rotation0)

    def _plt_get_text_fontfamily(self) -> str:
        return self._font_family

    def _plt_set_text_fontfamily(self, fontfamily: str):
        for child in self.get_children():
            child.set_fontfamily(fontfamily)
        self._font_family = fontfamily

    ##### HasFaces #####

    @staticmethod
    def _set_bbox_props(child: mplText, **bbox_props):
        if child.get_bbox_patch() is None:
            child.set_bbox(bbox_props)
        else:
            child.get_bbox_patch().update(bbox_props)

    def _plt_get_face_color(self):
        out = []
        for child in self.get_children():
            patch = child.get_bbox_patch()
            if patch is None:
                out.append(np.array([0, 0, 0, 0]))
            else:
                out.append(np.asarray(patch.get_facecolor()))
        return np.stack(out, axis=0)

    def _plt_set_face_color(self, color):
        color = as_color_array(color, len(self.get_children()))
        for child, color0 in zip(self.get_children(), color):
            self._set_bbox_props(child, facecolor=color0)

    def _plt_get_face_hatch(self) -> Hatch:
        out = []
        for child in self.get_children():
            patch = child.get_bbox_patch()
            if patch is None:
                out.append(Hatch.SOLID)
            else:
                out.append(Hatch(patch.get_hatch() or ""))
        return out

    def _plt_set_face_hatch(self, pattern: Hatch):
        if isinstance(pattern, Hatch):
            if pattern is Hatch.SOLID:
                ptn = [None] * len(self.get_children())
            else:
                ptn = [pattern.value] * len(self.get_children())
        else:
            ptn = [p.value if p is not Hatch.SOLID else None for p in pattern]
        for child, ptn0 in zip(self.get_children(), ptn):
            self._set_bbox_props(child, hatch=ptn0)

    def _plt_get_edge_color(self):
        out = []
        for child in self.get_children():
            patch = child.get_bbox_patch()
            if patch is None:
                out.append(np.array([0, 0, 0, 0]))
            else:
                out.append(np.asarray(patch.get_edgecolor()))
        return np.stack(out, axis=0)

    def _plt_set_edge_color(self, color):
        color = as_color_array(color, len(self.get_children()))
        for child, color0 in zip(self.get_children(), color):
            self._set_bbox_props(child, edgecolor=color0)

    def _plt_get_edge_width(self) -> float:
        out = []
        for child in self.get_children():
            patch = child.get_bbox_patch()
            if patch is None:
                out.append(0)
            else:
                out.append(patch.get_linewidth())
        return np.array(out, dtype=np.float32)

    def _plt_set_edge_width(self, width: float | NDArray[np.floating]):
        if is_real_number(width):
            width = np.full(len(self.get_children()), width, dtype=np.float32)
        for child, width0 in zip(self.get_children(), width):
            self._set_bbox_props(child, linewidth=width0)

    def _plt_get_edge_style(self) -> LineStyle:
        out = []
        for child in self.get_children():
            patch = child.get_bbox_patch()
            if patch is None:
                out.append(LineStyle.SOLID)
            else:
                style = patch.get_linestyle()
                if style == "solid":
                    out.append(LineStyle.SOLID)
                else:
                    out.append(LineStyle(style))
        return out

    def _plt_set_edge_style(self, style: LineStyle):
        if isinstance(style, LineStyle):
            if style is LineStyle.SOLID:
                sty = [None] * len(self.get_children())
            else:
                sty = [style.value] * len(self.get_children())
        else:
            sty = [s.value if s is not LineStyle.SOLID else None for s in style]
        for child, sty0 in zip(self.get_children(), sty):
            self._set_bbox_props(child, linestyle=sty0)


_VERTICAL_ALIGNMENTS = {
    "bottom": Alignment.BOTTOM,
    "baseline": Alignment.BOTTOM,
    "center": Alignment.CENTER,
    "center_baseline": Alignment.CENTER,
    "top": Alignment.TOP,
}
_VERTICAL_ALIGNMENTS_INV = {v: k for k, v in _VERTICAL_ALIGNMENTS.items()}
_HORIZONTAL_ALIGNMENTS = {
    "left": Alignment.LEFT,
    "center": Alignment.CENTER,
    "right": Alignment.RIGHT,
}
_HORIZONTAL_ALIGNMENTS_INV = {v: k for k, v in _HORIZONTAL_ALIGNMENTS.items()}


def _remove_method(this: Texts):
    for child in this.get_children():
        child.remove()



from __future__ import annotations

import bokeh.models as bk_models
import numpy as np
from numpy.typing import NDArray

from whitecanvas.backend.bokeh._base import (
    HeteroLayer,
    SupportsMouseEvents,
    from_bokeh_symbol,
    to_bokeh_symbol,
)
from whitecanvas.protocols import MarkersProtocol, check_protocol
from whitecanvas.types import Symbol
from whitecanvas.utils.type_check import is_real_number


@check_protocol(MarkersProtocol)
class Markers(HeteroLayer[bk_models.Scatter], SupportsMouseEvents):
    def __init__(self, xdata, ydata):
        ndata = len(xdata)
        self._data = bk_models.ColumnDataSource(
            data={
                "x": xdata,
                "y": ydata,
                "sizes": np.full(ndata, 10.0),
                "face_color": ["blue"] * ndata,
                "edge_color": ["black"] * ndata,
                "width": np.zeros(ndata),
                "pattern": [" "] * ndata,
                "style": ["solid"] * ndata,
                "hovertexts": [""] * ndata,
            }
        )
        self._model = bk_models.Scatter(
            x="x",
            y="y",
            size="sizes",
            line_color="edge_color",
            line_width="width",
            fill_color="face_color",
            hatch_pattern="pattern",
            line_dash="style",
        )
        self._visible = True

    def _plt_get_data(self):
        return self._data.data["x"], self._data.data["y"]

    def _plt_set_data(self, xdata: NDArray[np.number], ydata: NDArray[np.number]):
        ndata = self._data.data["x"].size
        cur_data = self._data.data.copy()
        cur_data["x"] = xdata
        cur_data["y"] = ydata
        cols_to_update = [
            "sizes", "face_color", "edge_color", "width", "pattern", "style",
            "hovertexts"
        ]  # fmt: skip
        if xdata.size < ndata:
            for key in cols_to_update:
                cur_data[key] = cur_data[key][: xdata.size]
        elif xdata.size > ndata:
            for key in cols_to_update:
                cur_data[key] = np.concatenate(
                    [cur_data[key], np.full(xdata.size - ndata, cur_data[key][-1])]
                )
        self._data.data = cur_data

    def _plt_get_symbol(self) -> Symbol:
        sym = self._model.marker
        rot = int(round(self._model.angle / np.pi * 2))
        return from_bokeh_symbol(sym, rot)

    def _plt_set_symbol(self, symbol: Symbol):
        sym, rot = to_bokeh_symbol(symbol)
        self._model.marker = sym
        self._model.angle = rot * np.pi / 2

    def _plt_get_symbol_size(self) -> NDArray[np.floating]:
        return self._data.data["sizes"]

    def _plt_set_symbol_size(self, size: float | NDArray[np.floating]):
        if is_real_number(size):
            size = np.full(self._plt_get_ndata(), size)
        self._data.data["sizes"] = size

    def _plt_get_ndata(self) -> int:
        return len(self._data.data["x"])



from whitecanvas import new_canvas


def main():
    canvas = new_canvas(backend="matplotlib:qt")
    x = [1, 2, 3, 4, 5]
    y0 = [2, 3, 2, 1, 2]
    y1 = [3, 2, 1, 2, 3]
    y2 = [0, 1, 2, 3, 4]
    b0 = canvas.add_bars(x, y0)
    b1 = canvas.stack_over(b0).add(y1)
    b2 = canvas.stack_over(b1).add(y2)
    print(f"stacked {b2} over {b1} over {b0}.")
    canvas.show(block=True)

if __name__ == "__main__":
    main()
